import copy
import itertools
import logging
from collections import defaultdict, deque, namedtuple
from pathlib import Path
from typing import Any, Callable, Optional, Union, cast
from warnings import warn

import pyarrow
from event_model import (
    DocumentNames,
    RunRouter,
    schema_validators,
    unpack_datum_page,
    unpack_event_page,
)
from event_model.documents import (
    Datum,
    DatumPage,
    DocumentType,
    Event,
    EventDescriptor,
    EventPage,
    Resource,
    RunStart,
    RunStop,
    StreamDatum,
    StreamResource,
)
from event_model.documents.event_descriptor import DataKey
from event_model.documents.stream_datum import StreamRange
from tiled.client import from_profile, from_uri
from tiled.client.base import BaseClient
from tiled.client.container import Container
from tiled.client.dataframe import DataFrameClient
from tiled.client.utils import handle_error
from tiled.structures.core import Spec
from tiled.utils import safe_json_dump

from ..consolidators import ConsolidatorBase, DataSource, StructureFamily, consolidator_factory
from ..run_engine import Dispatcher
from ..utils import truncate_json_overflow
from .core import CallbackBase
from .json_writer import JSONLinesWriter

# Aggregare the Event table rows and StreamDatums in batches before writing to Tiled
BATCH_SIZE = 10000

# Disallow using reserved words as data_keys identifiers
# Related: https://github.com/bluesky/event-model/pull/223
RESERVED_DATA_KEYS = ["time", "seq_num"]

# A lookup table for converting broad JSON types to numpy dtypes
JSON_TO_NUMPY_DTYPE = {"number": "<f8", "array": "<f8", "boolean": "|b1", "string": "<U0", "integer": "<i8"}

# A lookup table for converting Bluesky spec names to MIME types
MIMETYPE_LOOKUP = defaultdict(
    lambda: "application/octet-stream",
    {
        "hdf5": "application/x-hdf5",
        "AD_HDF5_SWMR_STREAM": "application/x-hdf5",
        "AD_HDF5_SWMR_SLICE": "application/x-hdf5",
        "PIL100k_HDF5": "application/x-hdf5",
        "XSP3": "application/x-hdf5",
        "XPS3": "application/x-hdf5",
        "XSP3_BULK": "application/x-hdf5",
        "XSP3_STEP": "application/x-hdf5",
        "AD_TIFF": "multipart/related;type=image/tiff",
        "AD_HDF5_GERM": "application/x-hdf5",
        "PIZZABOX_ENC_FILE_TXT_PD": "text/csv",
        "PANDA": "application/x-hdf5",
        "ROI_HDF5_FLY": "application/x-hdf5",
        "ROI_HDF51_FLY": "application/x-hdf5",
        "SIS_HDF51_FLY_STREAM_V1": "application/x-hdf5",
        "MERLIN_FLY_STREAM_V2": "application/x-hdf5",
        "MERLIN_HDF5_BULK": "application/x-hdf5",
        "TPX_HDF5": "application/x-hdf5",
        "EIGER2_STREAM": "application/x-hdf5",
        "NPY_SEQ": "multipart/related;type=application/x-npy",
        "ZEBRA_HDF51_FLY_STREAM_V1": "application/x-hdf5",
    },
)

logger = logging.getLogger(__name__)


def concatenate_stream_datums(*docs: StreamDatum):
    """Concatenate consecutive StreamDatum documents into a single StreamDatum document"""

    if len(docs) == 1:
        return docs[0]

    if len({doc["descriptor"] for doc in docs}) > 1:
        raise ValueError("All StreamDatum documents must reference the same descriptor.")
    if len({doc["stream_resource"] for doc in docs}) > 1:
        raise ValueError("All StreamDatum documents must reference the same stream_resource.")
    docs = tuple(sorted(docs, key=lambda doc: doc["indices"]["start"]))
    for d1, d2 in zip(docs[:-1], docs[1:]):  # TODO: use itertools.pairwise(docs) in python 3.10+
        if d1["indices"]["stop"] != d2["indices"]["start"]:
            raise ValueError("StreamDatum documents must be consecutive.")

    return StreamDatum(
        uid=docs[-1]["uid"],
        stream_resource=docs[-1]["stream_resource"],
        descriptor=docs[-1]["descriptor"],
        indices=StreamRange(start=docs[0]["indices"]["start"], stop=docs[-1]["indices"]["stop"]),
        seq_nums=StreamRange(start=docs[0]["seq_nums"]["start"], stop=docs[-1]["seq_nums"]["stop"]),
    )


# A named tuple to cache references to external data from Event documents.
ExternalEventDataReference = namedtuple(
    "ExternalEventDataReference",
    [
        "datum_id",  # The UID of the Datum document that references this external data
        "data_key",  # The data_key of the external data
        "desc_uid",  # The UID of the EventDescriptor document that this datum belongs to
        "seq_num",  # The sequence number of the Event document
    ],
)


class _ConditionalBackup:
    """Callback that tries to call the primary callback and, if it fails, flushes the buffer to backup callbacks.

    Once an error has been encountererd in the primary callback, all subsequent documents would be sent to the
    backup callbacks as well.

    This callback is intended to be used with a `RunRouter` and process documents from a single Bluesky run.
    """

    def __init__(self, primary_callback: Callable, backup_callbacks: list[Callable], maxlen: int = 1_000_000):
        self.primary_callback = primary_callback
        self.backup_callbacks = backup_callbacks
        self._buffer: deque[tuple[str, DocumentType]] = deque(maxlen=maxlen)
        self._push_to_backup = False

    def __call__(self, name: str, doc: DocumentType):
        self._buffer.append((name, doc))

        try:
            self.primary_callback(name, doc)
        except Exception as e:
            logger.warning(
                f"Primary callback {type(self.primary_callback).__name__} failed: {e}. "
                "Flushing buffer to backup callbacks.",
                stacklevel=2,
            )
            self._push_to_backup = True

        if self._push_to_backup:
            for name, doc in self._buffer:
                for bcb in self.backup_callbacks:
                    try:
                        bcb(name, doc)
                    except Exception as e:
                        logger.warning(
                            f"Backup callback {bcb.__class__.__name__} failed with error: {e}", stacklevel=2
                        )
            self._buffer.clear()


class RunNormalizer(CallbackBase):
    """Callback for updating Bluesky documents to their latest schema.

    This callback can be used to subscribe additional consumers that require the updated documents.
    Returns a shallow copy of the document to avoid modifying the original one.

    Parameters
    ----------
        patches : dict[str, Callable], optional
            A dictionary of patch functions to apply to the documents before modifying them.
            The keys are document names (e.g., "start", "stop", "descriptor", etc.), and the values
            are functions that take a document as input and return a modified document.
        spec_to_mimetype : dict[str, str], optional
            A dictionary mapping spec names to MIME types. This is used to convert `Resource` documents
            to the latest `StreamResource` schema.
            The supplied dictionary updates the default `MIMETYPE_LOOKUP` dictionary.
    """

    def __init__(
        self,
        patches: Optional[dict[str, Callable]] = None,
        spec_to_mimetype: Optional[dict[str, str]] = None,
    ):
        self._token_refs: dict[str, Callable] = {}
        self.dispatcher = Dispatcher()
        self.patches = patches or {}
        self.spec_to_mimetype = MIMETYPE_LOOKUP | (spec_to_mimetype or {})

        self._next_frame_index: dict[tuple[str, str], dict[str, int]] = defaultdict(
            lambda: {"carry": 0, "index": 0}
        )
        self._datum_cache: dict[str, Datum] = {}
        self._ext_ref_cache: list[ExternalEventDataReference] = []  # Cache for references to external Event data
        self._desc_name_by_uid: dict[str, str] = {}
        self._sres_cache: dict[str, StreamResource] = {}
        self._emitted: set[str] = set()  # UIDs of the StreamResource documents that have been emitted
        self._int_keys: set[str] = set()  # Names of internal data_keys
        self._ext_keys: set[str] = set()

    def _convert_resource_to_stream_resource(self, doc: Union[Resource, StreamResource]) -> StreamResource:
        """Make changes to and return a shallow copy of StreamRsource dictionary adhering to the new structure.

        Kept for back-compatibility with old StreamResource schema from event_model<1.20.0
        or Resource documents that are converted to StreamResources.
        """
        stream_resource_doc = cast(StreamResource, doc)

        if "mimetype" not in doc:
            # The document is a `Resource` or a < v1.20 `StreamResource`.
            # Both are converted to latest version `StreamResource`.
            for expected_key in ("spec", "root", "resource_path", "resource_kwargs"):
                if expected_key not in doc:
                    raise RuntimeError(
                        f"`Resource` or `StreamResource` legacy document is missing a '{expected_key}'"
                    )

            # Convert the Resource (or old StreamResource) document to a StreamResource document
            resource_dict = cast(dict, doc)
            stream_resource_doc["mimetype"] = self.spec_to_mimetype[resource_dict.pop("spec")]
            stream_resource_doc["parameters"] = resource_dict.pop("resource_kwargs", {})
            file_path = Path(resource_dict.pop("root").strip("/")).joinpath(
                resource_dict.pop("resource_path").strip("/")
            )
            stream_resource_doc["uri"] = "file://localhost/" + str(file_path).lstrip("/")

        # Ensure that the internal path within HDF5 files is referenced with "dataset" parameter
        if stream_resource_doc["mimetype"] == "application/x-hdf5":
            stream_resource_doc["parameters"]["dataset"] = stream_resource_doc["parameters"].pop(
                "path", stream_resource_doc["parameters"].pop("dataset", "")
            )

        # Ensure that only the necessary fields are present in the StreamResource document
        stream_resource_doc["data_key"] = stream_resource_doc.get("data_key", "")
        required_keys = {"data_key", "mimetype", "parameters", "uid", "uri"}
        for key in set(stream_resource_doc.keys()).difference(required_keys):
            stream_resource_doc.pop(key)  # type: ignore

        return stream_resource_doc

    def _convert_datum_to_stream_datum(
        self, datum_doc: Datum, data_key: str, desc_uid: str, seq_num: int
    ) -> tuple[Optional[StreamResource], StreamDatum]:
        """Convert the Datum document to the StreamDatum format

        This conversion requires (and is triggered when) the Event document is received. The function also returns
        a corresponding StreamResource document, if it hasn't been emitted yet.

        Parameters
        ----------
        datum_doc : Datum
            The Datum document to convert.
        data_key : str
            The data_key of the external data in the Event document; this parameter must be included in the new
            StreamResource document.
        desc_uid : str
            The UID of the EventDescriptor document that this datum belongs to.
        seq_num : int
            The sequence number of the Event document that this datum belongs to; 1-base index.

        Returns
        -------
        sres_doc : StreamResource, optional
            The corresponding StreamResource document, if it hasn't been emitted yet, otehrwise -- None.
        sdat_doc : StreamDatum
            The StreamDatum document corresponding to the Datum document.
        """

        # Some Datums contain datum_kwargs and the 'frame' field, which indicates the last index of the
        # frame. This should take precedence over the 'seq_num' field in the Event document. Keep the
        # last frame index in memory, since next Datums may refer to more than one frame (it is
        # assumed that Events always refer to a single frame).
        # There are cases when the frame_index is reset during the scan (e.g. if Datums for the same
        # data_key belong to different Resources), so the 'carry' field is used to keep track of the
        # previous frame index.
        datum_kwargs = datum_doc.get("datum_kwargs", {})
        frame = datum_kwargs.pop("frame", None)
        if frame is not None:
            desc_name = self._desc_name_by_uid[desc_uid]  # Name of the descriptor (stream)
            _next_index = self._next_frame_index[(desc_name, data_key)]
            index_start = sum(_next_index.values())
            _next_index["index"] = frame + 1
            index_stop = sum(_next_index.values())
            if index_stop < index_start:
                # The datum is likely referencing a next Resource, but the indexing must continue
                _next_index["carry"] = index_start
                index_stop = sum(_next_index.values())
        else:
            index_start, index_stop = seq_num - 1, seq_num
        indices = StreamRange(start=index_start, stop=index_stop)
        seq_nums = StreamRange(start=index_start + 1, stop=index_stop + 1)

        # produce the Resource document, if needed (add data_key to match the StreamResource schema)
        # Emit a copy of the StreamResource document with a new uid; this allows to account for cases
        # where one Resource is used by several data streams with different data_keys and datum_kwargs.
        sres_doc = None
        sres_uid = datum_doc["resource"]
        new_sres_uid = sres_uid + "-" + data_key
        if (sres_uid in self._sres_cache) and (new_sres_uid not in self._emitted):
            sres_doc = copy.deepcopy(self._sres_cache[sres_uid])
            sres_doc["data_key"] = data_key
            sres_doc["parameters"].update(datum_kwargs)
            sres_doc["uid"] = new_sres_uid

        # Produce the StreamDatum document
        sdat_doc = StreamDatum(
            uid=datum_doc["datum_id"],
            stream_resource=new_sres_uid,
            descriptor=desc_uid,
            indices=indices,
            seq_nums=seq_nums,
        )

        return sres_doc, sdat_doc

    def start(self, doc: RunStart):
        doc = copy.copy(doc)
        if patch := self.patches.get("start"):
            doc = patch(doc)
        self.emit(DocumentNames.start, doc)

    def stop(self, doc: RunStop):
        doc = copy.copy(doc)
        if patch := self.patches.get("stop"):
            doc = patch(doc)

        # If there are any cached references to external data, emit StreamResources and StreamDatums now
        for datum_id, data_key, desc_uid, seq_num in self._ext_ref_cache:
            if datum_doc := self._datum_cache.pop(datum_id, None):
                sres_doc, sdat_doc = self._convert_datum_to_stream_datum(datum_doc, data_key, desc_uid, seq_num)
                if (sres_doc is not None) and (sres_doc["uid"] not in self._emitted):
                    self.emit(DocumentNames.stream_resource, sres_doc)
                    self._emitted.add(sres_doc["uid"])
                self.emit(DocumentNames.stream_datum, sdat_doc)
            else:
                raise RuntimeError(
                    f"Cannot emit StreamDatum for {data_key} because the corresponding Datum document is missing."
                )

        self.emit(DocumentNames.stop, doc)

    def descriptor(self, doc: EventDescriptor):
        doc = copy.deepcopy(doc)
        if patch := self.patches.get("descriptor"):
            doc = patch(doc)

        # Rename data_keys that use reserved words, "time" and "seq_num"
        for name in RESERVED_DATA_KEYS:
            if name in doc["data_keys"].keys():
                if f"_{name}" in doc["data_keys"].keys():
                    raise ValueError(f"Cannot rename {name} to _{name} because it already exists")
                doc["data_keys"][f"_{name}"] = doc["data_keys"].pop(name)
                for obj_data_keys_list in doc["object_keys"].values():
                    if name in obj_data_keys_list:
                        obj_data_keys_list.remove(name)
                        obj_data_keys_list.append(f"_{name}")

        # Rename some fields (in-place) to match the current schema for the descriptor
        # Loop over all dictionaries that specify data_keys (both event data_keys or configuration data_keys)
        conf_data_keys = (obj["data_keys"].values() for obj in doc["configuration"].values())
        for data_keys_spec in itertools.chain(doc["data_keys"].values(), *conf_data_keys):
            # Determine numpy data type. From highest precedent to lowest:
            # 1. Try 'dtype_descr', optional, if present -- this is a structural dtype
            # 2. Try 'dtype_numpy', optional in the document schema.
            # 3. Try 'dtype_str', an old convention predataing 'dtype_numpy', not in the schema.
            # 4. Get 'dtype', required by the schema, which is a fuzzy JSON spec like 'number'
            #    and make a best effort to convert it to a numpy spec like '<u8'.
            # 5. If unable to do any of the above, do not set 'dtype_numpy' at all.
            dtype_descr = data_keys_spec.pop("dtype_descr", [])  # type: ignore
            dtype_str = data_keys_spec.pop("dtype_str", None)  # type: ignore
            if dtype_numpy := (
                list(map(list, dtype_descr))
                or data_keys_spec.get("dtype_numpy", dtype_str)
                or JSON_TO_NUMPY_DTYPE.get(data_keys_spec["dtype"])
            ):
                data_keys_spec["dtype_numpy"] = dtype_numpy

        # Ensure that all event data_keys have object_name assigned (for consistency)
        for obj_name, data_keys_list in doc["object_keys"].items():
            for key in data_keys_list:
                doc["data_keys"][key]["object_name"] = obj_name

        # Keep names of external and internal data_keys
        data_keys = doc.get("data_keys", {})
        self._int_keys.update({k for k, v in data_keys.items() if "external" not in v.keys()})
        self._ext_keys.update({k for k, v in data_keys.items() if "external" in v.keys()})
        for key in self._ext_keys:
            if key in data_keys:
                data_keys[key]["external"] = data_keys[key].pop("external", "")  # Make sure the value is not None

        # Keep a reference to the descriptor name (stream) by its uid
        self._desc_name_by_uid[doc["uid"]] = doc["name"]

        # Emit the updated descriptor document
        self.emit(DocumentNames.descriptor, doc)

    def event(self, doc: Event):
        doc = copy.deepcopy(doc)
        if patch := self.patches.get("event"):
            doc = patch(doc)

        # Part 0. ----- Preprocessing -----
        # Rename data_keys that use reserved words, "time" and "seq_num"
        for name in RESERVED_DATA_KEYS:
            if name in doc["data"].keys():
                doc["data"][f"_{name}"] = doc["data"].pop(name)
                doc["timestamps"][f"_{name}"] = doc["timestamps"].pop(name)
            if name in doc["filled"].keys():
                doc["filled"][f"_{name}"] = doc["filled"].pop(name)

        # Part 1. ----- Internal Data -----
        # Emit a new Event with _internal_ data: select only keys without 'external' flag or those that are filled
        filled = doc.pop("filled", {})
        event_keys = [k for k in self._int_keys if filled.get(k, True)] + [
            k for k in self._ext_keys if filled.get(k, False)
        ]
        event_doc = copy.copy(doc)  # Keep another copy with the external data_keys intact
        event_doc["data"] = {k: v for k, v in doc["data"].items() if k in event_keys}
        event_doc["timestamps"] = {k: v for k, v in doc["timestamps"].items() if k in event_keys}
        self.emit(DocumentNames.event, event_doc)

        # Part 2. ----- External Data -----
        # Process _external_ data: Loop over all referenced Datums and all external data keys that are not filled
        for data_key, datum_id in doc["data"].items():
            if data_key not in set(self._ext_keys).difference(event_keys):
                continue  # Skip internal data_keys
            if datum_doc := self._datum_cache.pop(datum_id, None):
                sres_doc, sdat_doc = self._convert_datum_to_stream_datum(
                    datum_doc, data_key, desc_uid=doc["descriptor"], seq_num=doc["seq_num"]
                )
                if (sres_doc is not None) and (sres_doc["uid"] not in self._emitted):
                    self.emit(DocumentNames.stream_resource, sres_doc)
                    self._emitted.add(sres_doc["uid"])  # Mark the StreamResource as emitted
                self.emit(DocumentNames.stream_datum, sdat_doc)
            else:
                # This Event references a Datum that has not been received yet; cache and process it later
                missing = ExternalEventDataReference(datum_id, data_key, doc["descriptor"], doc["seq_num"])
                self._ext_ref_cache.append(missing)

    def resource(self, doc: Resource):
        doc = copy.copy(doc)
        if patch := self.patches.get("resource"):
            doc = patch(doc)

        # Convert the Resource document to StreamResource format
        self._sres_cache[doc["uid"]] = self._convert_resource_to_stream_resource(doc)

    def stream_resource(self, doc: StreamResource):
        doc = copy.copy(doc)
        if patch := self.patches.get("stream_resource"):
            doc = patch(doc)

        # Convert the StreamResource document to the latest schema
        doc = self._convert_resource_to_stream_resource(doc)
        self.emit(DocumentNames.stream_resource, doc)

    def stream_datum(self, doc: StreamDatum):
        doc = copy.copy(doc)
        if patch := self.patches.get("stream_datum"):
            doc = patch(doc)
        self.emit(DocumentNames.stream_datum, doc)

    def datum(self, doc: Datum):
        doc = copy.copy(doc)
        if patch := self.patches.get("datum"):
            doc = patch(doc)

        self._datum_cache[doc["datum_id"]] = doc

    def datum_page(self, doc: DatumPage):
        for _doc in unpack_datum_page(doc):
            self.datum(_doc)

    def event_page(self, doc: EventPage):
        for _doc in unpack_event_page(doc):
            self.event(_doc)

    def emit(self, name, doc):
        """Check the document schema and send to the dispatcher"""
        schema_validators[name].validate(doc)
        self.dispatcher.process(name, doc)

    def subscribe(self, func, name="all"):
        """Convenience function for dispatcher subscription"""
        token = self.dispatcher.subscribe(func, name)
        self._token_refs[token] = func
        return token

    def unsubscribe(self, token):
        """Convenience function for dispatcher un-subscription"""
        self._token_refs.pop(token, None)
        self.dispatcher.unsubscribe(token)


class _RunWriter(CallbackBase):
    """Write documents from a single Bluesky Run into Tiled.

    This callback is intended to be used with a `RunRouter` and process documents from a single Bluesky run.
    It creates a new Tiled Container for the run and writes the internal data provided in Event documents
    as well as registers external files provided in StreamResource and StreamDatum documents.

    The callback is intended to be used with the most recent version of EventModel; to support legacy
    schemas, use the `RunNormalizer` callback first to update the documents before writing them to Tiled.

    Parameters
    ----------
        client : BaseClient
            The Tiled client to use for writing the data.
    """

    def __init__(self, client: BaseClient, batch_size: int = BATCH_SIZE):
        self.client = client
        self.root_node: Union[None, Container] = None
        self._desc_nodes: dict[str, Container] = {}  # references to the descriptor nodes by their uid's and names
        self._sres_nodes: dict[str, BaseClient] = {}
        self._internal_tables: dict[str, DataFrameClient] = {}  # references to the internal tables by desc_names
        self._stream_resource_cache: dict[str, StreamResource] = {}
        self._consolidators: dict[str, ConsolidatorBase] = {}
        self._internal_data_cache: dict[str, list[dict[str, Any]]] = defaultdict(list)
        self._external_data_cache: dict[str, StreamDatum] = {}  # sres_uid : (concatenated) StreamDatum
        self._batch_size = batch_size
        self.data_keys: dict[str, DataKey] = {}
        self.access_tags = None

    def _write_internal_data(self, data_cache: list[dict[str, Any]], desc_node: Container):
        """Write the internal data table to Tiled and clear the cache."""

        desc_name = desc_node.item["id"]  # Name of the descriptor (stream)
        table = pyarrow.Table.from_pylist(data_cache)

        if not (df_client := self._internal_tables.get(desc_name)):
            # Create a new "internal" data node and write the initial piece of data
            metadata = {k: v for k, v in self.data_keys.items() if k in table.column_names}
            metadata = truncate_json_overflow(metadata)
            # Replace any nulls in the schema with string type
            schema = copy.copy(table.schema)
            for i, field in enumerate(table.schema):
                if pyarrow.types.is_null(field.type):
                    schema = schema.set(i, field.with_type(pyarrow.string()))
                elif pyarrow.types.is_list(field.type) and pyarrow.types.is_null(field.type.value_type):
                    schema = schema.set(i, field.with_type(pyarrow.list_(pyarrow.string())))
            # Initialize the table and keep a reference to the client
            df_client = desc_node.create_appendable_table(
                schema=schema, key="internal", metadata=metadata, access_tags=self.access_tags
            )
            self._internal_tables[desc_name] = df_client

        df_client.append_partition(table, 0)

    def _write_external_data(self, doc: StreamDatum):
        """Register the external data provided in StreamDatum in Tiled"""

        sres_uid, desc_uid = doc["stream_resource"], doc["descriptor"]
        sres_node, consolidator = self.get_sres_node(sres_uid, desc_uid)
        consolidator.consume_stream_datum(doc)
        self._update_data_source_for_node(sres_node, consolidator.get_data_source())

    def _update_data_source_for_node(self, node: BaseClient, data_source: DataSource):
        """Update StreamResource node in Tiled"""
        data_source.id = node.data_sources()[0].id  # ID of the existing DataSource record
        handle_error(
            node.context.http_client.put(
                node.uri.replace("/metadata/", "/data_source/", 1),
                content=safe_json_dump({"data_source": data_source}),
            )
        ).json()

    def start(self, doc: RunStart):
        doc = copy.copy(doc)
        self.access_tags = doc.pop("tiled_access_tags", None)  # type: ignore
        self.root_node = self.client.create_container(
            key=doc["uid"],
            metadata={"start": truncate_json_overflow(dict(doc))},
            specs=[Spec("BlueskyRun", version="3.0")],
            access_tags=self.access_tags,
        )

    def stop(self, doc: RunStop):
        if self.root_node is None:
            raise RuntimeError("RunWriter is not properly initialized: no Start document has been recorded.")

        # Write the cached internal data
        for desc_name, data_cache in self._internal_data_cache.items():
            if data_cache:
                self._write_internal_data(data_cache, desc_node=self._desc_nodes[desc_name])
                data_cache.clear()

        # Write the cached StreamDatums data
        for stream_datum_doc in self._external_data_cache.values():
            self._write_external_data(stream_datum_doc)

        # Validate structure for some StreamResource nodes
        for sres_uid, sres_node in self._sres_nodes.items():
            consolidator = self._consolidators[sres_uid]
            if consolidator._sres_parameters.get("_validate", False):
                try:
                    consolidator.validate(fix_errors=True)
                except Exception as e:
                    msg = f"{type(e).__name__}: " + str(e).replace("\n", " ").replace("\r", "").strip()
                    warn(f"Validation of StreamResource {sres_uid} failed with error: {msg}", stacklevel=2)
                self._update_data_source_for_node(sres_node, consolidator.get_data_source())

        # Write the stop document to the metadata
        self.root_node.update_metadata(metadata={"stop": doc, **dict(self.root_node.metadata)}, drop_revision=True)

    def descriptor(self, doc: EventDescriptor):
        if self.root_node is None:
            raise RuntimeError("RunWriter is not properly initialized: no Start document has been recorded.")

        desc_name = doc["name"]  # Name of the descriptor/stream
        self.data_keys.update(doc.get("data_keys", {}))

        # Create a new Container with "composite" spec for the stream if it does not exist
        # Since the data_keys are guaranteed to be unique, we don't need to perform client-side validation of the
        # "composite" spec constraints and can use the `.base` (Container) client directly.
        if desc_name not in self._desc_nodes.keys():
            metadata = {k: v for k, v in doc.items() if k not in {"name", "object_keys", "run_start"}}
            desc_node = self.root_node.create_container(
                key=desc_name,
                metadata=truncate_json_overflow(metadata),
                specs=[Spec("BlueskyEventStream", version="3.0"), Spec("composite")],
                access_tags=self.access_tags,
            ).base
        else:
            # Rare Case: This new descriptor likely updates stream configs mid-experiment
            # We assume tha the full descriptor has been already received, so we don't need to store everything
            # but only the uid, timestamp, and also data and timestamps in configuration (without conf specs).
            desc_node = self._desc_nodes[desc_name]
            updates = desc_node.metadata.get("_config_updates", []) + [{"uid": doc["uid"], "time": doc["time"]}]
            if conf_meta := doc.get("configuration"):
                updates[-1].update({"configuration": conf_meta})
            # Update the metadata with the new configuration
            metadata = {"_config_updates": truncate_json_overflow(updates)}
            desc_node.update_metadata(metadata=metadata, drop_revision=True)

        self._desc_nodes[doc["uid"]] = self._desc_nodes[desc_name] = desc_node  # Keep a reference to the node

    def event(self, doc: Event):
        desc_uid = doc["descriptor"]
        desc_name = self._desc_nodes[desc_uid].item["id"]  # Name of the descriptor (stream)

        # Do not write the data immediately; collect it in a cache and write in bulk later
        data_cache = self._internal_data_cache[desc_name]
        row = {"seq_num": doc["seq_num"], "time": doc["time"], **doc["data"]}
        row.update({f"ts_{k}": v for k, v in doc["timestamps"].items()})
        data_cache.append(row)

        if len(data_cache) >= self._batch_size:
            self._write_internal_data(data_cache, desc_node=self._desc_nodes[desc_uid])
            data_cache.clear()

    def event_page(self, doc: EventPage):
        for _doc in unpack_event_page(doc):
            self.event(_doc)

    def stream_resource(self, doc: StreamResource):
        self._stream_resource_cache[doc["uid"]] = doc

    def get_sres_node(self, sres_uid: str, desc_uid: Optional[str] = None) -> tuple[BaseClient, ConsolidatorBase]:
        """Get the Tiled node and the associate Consolidator corresponding to the data_key in StreamResource

        If the node does not exist, register it from a cached StreamResource document. Keep a reference to the
        node and the corresponding Consolidator object. If the node already exists, return the existing one.

        The nodes and consolidators are referenced by both:
        - sres_uid: the uid of the StreamResource document
        - desc_name + data_key: the name of the descriptor (stream) and the data_key
        """

        if sres_uid in self._sres_nodes.keys():
            sres_node = self._sres_nodes[sres_uid]
            consolidator = self._consolidators[sres_uid]

        elif sres_uid in self._stream_resource_cache.keys():
            if not desc_uid:
                raise RuntimeError("Descriptor uid must be specified to initialise a Stream Resource node")

            sres_doc = self._stream_resource_cache[sres_uid]
            desc_node = self._desc_nodes[desc_uid]
            full_data_key = f"{desc_node.item['id']}_{sres_doc['data_key']}"  # desc_name + data_key

            # Check if there already exists a Node and a Consolidator for this data_key
            # i.e. this is an additional StreamResource, whose data should be concatenated with the existing one
            if full_data_key in self._sres_nodes.keys():
                sres_node = self._sres_nodes[full_data_key]
                consolidator = self._consolidators[full_data_key]
                consolidator.update_from_stream_resource(sres_doc)
            else:
                consolidator = consolidator_factory(sres_doc, desc_node.metadata)
                sres_node = desc_node.new(
                    key=consolidator.data_key,
                    structure_family=StructureFamily.array,
                    data_sources=[consolidator.get_data_source()],
                    metadata={},
                    specs=[],
                    access_tags=self.access_tags,
                )

            self._consolidators[sres_uid] = self._consolidators[full_data_key] = consolidator
            self._sres_nodes[sres_uid] = self._sres_nodes[full_data_key] = sres_node
        else:
            raise RuntimeError(f"Stream Resource {sres_uid} is referenced before being received.")

        return sres_node, consolidator

    def stream_datum(self, doc: StreamDatum):
        if self._batch_size <= 1:
            # If batch size is 1, write the StreamDatum immediately
            self._write_external_data(doc)
            return

        # Try to concatenate and cache the StreamDatum document to process it later
        sres_uid = doc["stream_resource"]
        if cached_stream_datum_doc := self._external_data_cache.pop(sres_uid, None):
            try:
                _doc = concatenate_stream_datums(cached_stream_datum_doc, doc)
                if _doc["indices"]["stop"] - _doc["indices"]["start"] >= self._batch_size:
                    self._write_external_data(_doc)
                else:
                    self._external_data_cache[sres_uid] = _doc
            except ValueError:
                # If concatenation fails, write the cached document and then the new one immediately
                self._write_external_data(cached_stream_datum_doc)
                self._write_external_data(doc)
        else:
            self._external_data_cache[sres_uid] = doc


class TiledWriter:
    """Callback for write metadata and data from Bluesky documents into Tiled.

    This callback relies on the `RunRouter` to route documents from one or more runs into
    independent instances of the `_RunWriter` callback. The `RunRouter` is responsible for
    creating a new instance of the `_RunWriter` for each run.

    Parameters
    ----------
        client : `tiled.client.BaseClient`
            The Tiled client to use for writing data. This client must be initialized with
            the appropriate credentials and connection parameters to access the Tiled server.
        normalizer : Optional[CallbackBase]
            A callback for normalizing Bluesky documents to the latest schema. If not provided,
            the default `RunNormalizer` will be used. The supplied normalizer should accept
            `patches` and `spec_to_mimetype` (or `**kwargs`) for initialization.
            To disable normalization and pass the incoming document directly to _RunWriter,
            set this parameter to `None`.
        patches : Optional[dict[str, Callable]]
            A dictionary of patch functions to apply to specific document types before normalizing
            and writing them. The keys should be the document names (e.g., "start", "stop",
            "descriptor", etc.), and the values should be functions that take a document and return
            a modified document of the same type.
            This argument is ignored if `normalizer` is set to `None`.
        spec_to_mimetype : Optional[dict[str, str]]
            A dictionary mapping spec names to MIME types. This is used to convert `Resource` documents
            to the latest `StreamResource` schema. If not provided, the default mapping will be used.
            This argument is ignored if `normalizer` is set to `None`.
        backup_directory : Optional[str]
            If specified, this directory will be used to back up runs that fail to be written
            to Tiled. All documents for the entire Bluesky Run will be written in JSONLines format,
            allowing for recovery in case of errors during the writing process.
        batch_size : int
            The number of Events or StreamDatums collect before writing them to Tiled.
            This is useful for reducing the number of write operations and improving performance when
            writing large amounts of data (e.g. database migration). For streaming applications,
            it is recommended to set this parameter to <= 1, so that each Event or StreamDatum is written
            to Tiled immediately after they are received.
    """

    def __init__(
        self,
        client: BaseClient,
        *,
        normalizer: Optional[type[CallbackBase]] = RunNormalizer,
        patches: Optional[dict[str, Callable]] = None,
        spec_to_mimetype: Optional[dict[str, str]] = None,
        backup_directory: Optional[str] = None,
        batch_size: int = BATCH_SIZE,
    ):
        self.client = client.include_data_sources()
        self.patches = patches or {}
        self.spec_to_mimetype = spec_to_mimetype or {}
        self.backup_directory = backup_directory
        self._normalizer = normalizer
        self._run_router = RunRouter([self._factory])
        self._batch_size = batch_size

    def _factory(self, name, doc):
        """Factory method to create a callback for writing a single run into Tiled."""
        cb = run_writer = _RunWriter(self.client, batch_size=self._batch_size)

        if self._normalizer:
            # If normalize is True, create a RunNormalizer callback to update documents to the latest schema
            cb = self._normalizer(patches=self.patches, spec_to_mimetype=self.spec_to_mimetype)
            cb.subscribe(run_writer)

        if self.backup_directory:
            # If backup_directory is specified, create a conditional backup callback writing documents to JSONLines
            cb = _ConditionalBackup(cb, [JSONLinesWriter(self.backup_directory)])

        return [cb], []

    @classmethod
    def from_uri(
        cls,
        uri,
        *,
        normalizer: Optional[type[CallbackBase]] = RunNormalizer,
        patches: Optional[dict[str, Callable]] = None,
        spec_to_mimetype: Optional[dict[str, str]] = None,
        backup_directory: Optional[str] = None,
        batch_size: int = BATCH_SIZE,
        **kwargs,
    ):
        client = from_uri(uri, **kwargs)
        return cls(
            client,
            normalizer=normalizer,
            patches=patches,
            spec_to_mimetype=spec_to_mimetype,
            backup_directory=backup_directory,
            batch_size=batch_size,
        )

    @classmethod
    def from_profile(
        cls,
        profile,
        *,
        normalizer: Optional[type[CallbackBase]] = RunNormalizer,
        patches: Optional[dict[str, Callable]] = None,
        spec_to_mimetype: Optional[dict[str, str]] = None,
        backup_directory: Optional[str] = None,
        batch_size: int = BATCH_SIZE,
        **kwargs,
    ):
        client = from_profile(profile, **kwargs)
        return cls(
            client,
            normalizer=normalizer,
            patches=patches,
            spec_to_mimetype=spec_to_mimetype,
            backup_directory=backup_directory,
            batch_size=batch_size,
        )

    def __call__(self, name, doc):
        self._run_router(name, doc)



# Example: nested runs

from databroker import Broker
from event_model import RunRouter
from ophyd.sim import hw

import bluesky.plan_stubs as bps
import bluesky.preprocessors as bpp
from bluesky import RunEngine
from bluesky.callbacks.best_effort import BestEffortCallback

hw = hw()

RE = RunEngine({})

db = Broker.named("temp")
RE.subscribe(db.insert)


def factory(name, doc):
    # Documents from each run is routed to an independent
    #   instance of BestEffortCallback
    bec = BestEffortCallback()
    return [bec], []


rr = RunRouter([factory])
RE.subscribe(rr)


@bpp.set_run_key_decorator("run_2")
@bpp.run_decorator(md={})
def sim_plan_inner(npts):
    yield from bps.declare_stream(hw.motor1, hw.motor2, hw.det2, name="primary")
    for j in range(npts):
        yield from bps.mov(hw.motor1, j * 0.1 + 1, hw.motor2, j * 0.2 - 2)
        yield from bps.trigger_and_read([hw.motor1, hw.motor2, hw.det2])


@bpp.set_run_key_decorator("run_1")
@bpp.run_decorator(md={})
def sim_plan_outer(npts):
    yield from bps.declare_stream(hw.motor, hw.det, name="primary")
    for j in range(int(npts / 2)):
        yield from bps.mov(hw.motor, j * 0.2)
        yield from bps.trigger_and_read([hw.motor, hw.det])

    yield from sim_plan_inner(npts + 1)

    for j in range(int(npts / 2), npts):
        yield from bps.mov(hw.motor, j * 0.2)
        yield from bps.trigger_and_read([hw.motor, hw.det])



import time
from contextlib import contextmanager

import pytest

from bluesky.callbacks import CallbackBase
from bluesky.callbacks.buffer import BufferingWrapper


class SlowDummyCallback(CallbackBase):
    """Simulates a slow (blocking) callback for testing threaded wrappers."""

    def __init__(self, delay=0.1):
        self.delay = delay
        self.called = []

    def __call__(self, name, doc):
        time.sleep(self.delay)  # Simulate slow work
        self.called.append((name, doc))


class CallbackWithException(CallbackBase):
    """A callback that raises an exception when called."""

    def __call__(self, name, doc):
        raise RuntimeError("This callback always raises an exception.")


@pytest.fixture
def fast_cb():
    yield SlowDummyCallback(delay=0)


@pytest.fixture
def slow_cb():
    yield SlowDummyCallback(delay=0.1)


@contextmanager
def wait_for_condition(condition, timeout=3, interval=0.01):
    """Wait for a condition to become True within a timeout period.

    Parameters
    ----------
        condition : callable
            A function that returns True when the condition is met.
        timeout : float
            Maximum time to wait for the condition to become True.
        interval : float, optional
            Time to wait between checks of the condition (default is 0.01 seconds).

    Usage
    -----
        with wait_for_condition(1, lambda: x > y):
            pass
    """
    start = time.time()
    while time.time() - start < timeout:
        if condition():
            yield
            return
        time.sleep(interval)
    raise TimeoutError("Condition not met within timeout")


@pytest.mark.parametrize("cb", ["fast_cb", "slow_cb"])
def test_calls_are_delegated(cb, request):
    cb = request.getfixturevalue(cb)
    buff_cb = BufferingWrapper(cb)

    buff_cb("start", {"x": 1})
    with wait_for_condition(lambda: ("start", {"x": 1}) in cb.called):
        assert len(cb.called) == 1

    buff_cb("stop", {"x": 2})
    with wait_for_condition(lambda: ("stop", {"x": 2}) in cb.called):
        assert len(cb.called) == 2


@pytest.mark.parametrize("cb", ["fast_cb", "slow_cb"])
def test_calls_are_delegated_and_finished(cb, request):
    cb = request.getfixturevalue(cb)
    buff_cb = BufferingWrapper(cb)

    assert buff_cb._thread.is_alive()
    assert len(cb.called) == 0

    buff_cb("start", {"x": 1})
    buff_cb("stop", {"x": 2})

    assert buff_cb._thread.is_alive()
    buff_cb.shutdown()
    assert not buff_cb._thread.is_alive()

    assert ("start", {"x": 1}) in cb.called
    assert ("stop", {"x": 2}) in cb.called


@pytest.mark.parametrize("cb", ["fast_cb", "slow_cb"])
def test_graceful_shutdown_blocks_queue(cb, request):
    cb = request.getfixturevalue(cb)
    buff_cb = BufferingWrapper(cb)

    buff_cb("event", {"data": 42})
    buff_cb.shutdown()

    with pytest.raises(RuntimeError):
        buff_cb("post-shutdown", {"fail": True})

    assert ("event", {"data": 42}) in cb.called
    assert ("post-shutdown", {"fail": True}) not in cb.called


@pytest.mark.parametrize("cb", ["fast_cb", "slow_cb"])
def test_double_shutdown_does_not_fail(cb, request):
    cb = request.getfixturevalue(cb)
    buff_cb = BufferingWrapper(cb)

    buff_cb("one", {})
    buff_cb.shutdown()
    buff_cb.shutdown()  # Second shutdown should be a no-op

    assert ("one", {}) in cb.called


@pytest.mark.parametrize("cb, expected_min_duration", [("fast_cb", 0.0), ("slow_cb", 0.5)])
def test_shutdown_waits_for_processing(cb, expected_min_duration, request):
    cb = request.getfixturevalue(cb)
    buff_cb = BufferingWrapper(cb)

    for i in range(5):
        buff_cb("event", {"val": i})

    t0 = time.time()
    time.sleep(0.02)  # Let the last document enter the queue before shutdown
    buff_cb.shutdown()
    duration = time.time() - t0
    assert duration >= expected_min_duration  # Ensure shutdown waited for processing

    # After shutdown, all should be processed
    assert len(cb.called) == 5
    for i in range(5):
        assert ("event", {"val": i}) in cb.called


@pytest.mark.parametrize("cb, expected_max_duration", [("fast_cb", 0.1), ("slow_cb", 0.1)])
def test_shutdown_without_wait(cb, expected_max_duration, request):
    cb = request.getfixturevalue(cb)
    buff_cb = BufferingWrapper(cb)

    for i in range(5):
        buff_cb("event", {"val": i})

    # Shutdown without waiting — thread may still be running briefly
    t0 = time.time()
    buff_cb.shutdown(wait=False)
    duration = time.time() - t0
    assert duration < expected_max_duration
    assert len(cb.called) <= 5


@pytest.mark.parametrize("cb", ["fast_cb", "slow_cb"])
def test_shutdown_stops_processing_new_items(cb, request):
    cb = request.getfixturevalue(cb)
    buff_cb = BufferingWrapper(cb)

    buff_cb("one", {})
    buff_cb.shutdown()

    with pytest.raises(RuntimeError):
        buff_cb("two", {})

    assert ("one", {}) in cb.called
    assert ("two", {}) not in cb.called


def test_execption_in_callback():
    """Test that exceptions in the callback are handled gracefully."""
    cb = CallbackWithException()
    buff_cb = BufferingWrapper(cb)

    with pytest.raises(RuntimeError, match="This callback always raises an exception."):
        cb("test", {"data": 123})

    # Ensure that the exception does not crash the buffering wrapper
    buff_cb("test", {"data": 123})
    assert buff_cb._thread.is_alive()  # The thread should still be running
    buff_cb("test", {"data": 123})


def test_callback_logging_exceptions(monkeypatch):
    from types import SimpleNamespace
    from unittest.mock import MagicMock

    logger = SimpleNamespace(exception=MagicMock())
    monkeypatch.setattr("bluesky.callbacks.buffer.logger", logger)

    cb = CallbackWithException()
    buff_cb = BufferingWrapper(cb)

    assert logger.exception.call_count == 0
    buff_cb("test", {"data": 123})
    with wait_for_condition(lambda: logger.exception.call_count == 1):
        assert True



# declare module content to calm pyflakes about unused names
__all__ = [
    "CallbackBase",
    "CallbackCounter",
    "print_metadata",
    "collector",
    "get_obj_fields",
    "CollectThenCompute",
    "LiveTable",
    "LiveFit",
    "LiveScatter",
    "LivePlot",
    "LiveGrid",
    "LiveFitPlot",
    "LiveRaster",
    "LiveMesh",
]

from .core import (
    CallbackBase,
    CallbackCounter,
    CollectThenCompute,
    LiveTable,
    collector,
    get_obj_fields,
    print_metadata,
)
from .fitting import LiveFit
from .mpl_plotting import LiveFitPlot, LiveGrid, LiveMesh, LivePlot, LiveRaster, LiveScatter



"""Utilities for core."""

from enum import Enum, unique
from functools import wraps

from .errors import TransitionError


def is_(self, state):
    """Check if machine is in given state."""
    translator = self._meta['translator']
    state = translator.translate(state)
    return self.actual_state == state


def can_be_(self, state):
    """Check if machine can transit to given state."""
    translator = self._meta['translator']
    state = translator.translate(state)

    if self._meta['complete']:
        return True

    if self.actual_state is None:
        return True

    transitions = self._meta['transitions'][self.actual_state]
    return state in transitions


def force_set(self, state):
    """Set new state without checking if transition is allowed."""
    translator = self._meta['translator']
    state = translator.translate(state)
    attr = self._meta['state_attribute_name']
    setattr(self, attr, state)


def set_(self, state):
    """Set new state for machine."""
    if not self.can_be_(state):
        state = self._meta['translator'].translate(state)
        raise TransitionError(
            "Cannot transit from '{actual_value}' to '{value}'."
            .format(actual_value=self.actual_state.value, value=state.value)
        )

    self.force_set(state)


def state_getter(self):
    """Get actual state as value."""
    try:
        return self.actual_state.value
    except AttributeError:
        return None


def state_setter(self, value):
    """Set new state for machine."""
    self.set_(value)


def generate_getter(value):
    """Generate getter for given value."""
    @property
    @wraps(is_)
    def getter(self):
        return self.is_(value)

    return getter


def generate_checker(value):
    """Generate state checker for given value."""
    @property
    @wraps(can_be_)
    def checker(self):
        return self.can_be_(value)

    return checker


def generate_setter(value):
    """Generate setter for given value."""
    @wraps(set_)
    def setter(self):
        self.set_(value)

    return setter

state_property = property(state_getter, state_setter)


@property
def actual_state(self):
    """Actual state as `None` or `enum` instance."""
    attr = self._meta['state_attribute_name']
    return getattr(self, attr)


@property
def as_enum(self):
    """Return actual state as enum."""
    return self.actual_state


class EnumValueTranslator(object):

    """Helps to find enum element by its value."""

    def __init__(self, base_enum):
        """Init.

        :param enum base_enum: Enum, to which elements values are translated.

        """
        base_enum = unique(base_enum)
        self.base_enum = base_enum
        self.generate_search_table()

    def generate_search_table(self):
        self.search_table = dict(
            (item.value, item) for item in list(self.base_enum)
        )

    def translate(self, value):
        """Translate value to enum instance.

        If value is already enum instance, check if this value belongs to base
        enum.

        """
        if self._check_if_already_proper(value):
            return value

        try:
            return self.search_table[value]
        except KeyError:
            raise ValueError("Value {value} doesn't match any state.".format(
                value=value
            ))

    def _check_if_already_proper(self, value):
        if isinstance(value, Enum):
            if value in self.base_enum:
                return True
            raise ValueError(
                "Given value ('{value}') doesn't belong to states enum."
                .format(value=value)
            )
        return False



from opentelemetry import trace
from opentelemetry.sdk.resources import SERVICE_NAME, Resource
from opentelemetry.sdk.trace import TracerProvider
from opentelemetry.sdk.trace.export import BatchSpanProcessor, ConsoleSpanExporter
from ophyd.sim import det1

import bluesky.plan_stubs as bps
import bluesky.preprocessors as bpp
from bluesky.run_engine import RunEngine
from bluesky.tracing import trace_plan, tracer

# Service name is required for most backends,
# and although it's not necessary for console export,
# it's good to set service name anyways.
resource = Resource(attributes={SERVICE_NAME: "bluesky-docs-example"})

traceProvider = TracerProvider(resource=resource)
processor = BatchSpanProcessor(ConsoleSpanExporter())
traceProvider.add_span_processor(processor)
trace.set_tracer_provider(traceProvider)

RE = RunEngine()


@bpp.run_decorator()
@trace_plan(tracer, "demo plan")
def test_tracing_plan():
    yield from bps.abs_set(det1._motor, 5)


RE(test_tracing_plan())



from abc import abstractmethod
from collections.abc import AsyncIterator, Awaitable, Iterator
from typing import (
    Any,
    Callable,
    Generic,
    Literal,
    Optional,
    Protocol,
    TypeVar,
    Union,
    runtime_checkable,
)

from event_model.documents import Datum, StreamDatum, StreamResource
from event_model.documents.event import PartialEvent

# Including Dtype here because ophyd imports Dtype directly from protocols, not event-model.
from event_model.documents.event_descriptor import DataKey, Dtype
from event_model.documents.event_page import PartialEventPage
from event_model.documents.resource import PartialResource
from typing_extensions import ParamSpec, TypedDict, Unpack

# Squashes warning
Dtype = Dtype  # type: ignore


# TODO: these are not placed in Events by RE yet
class ReadingOptional(TypedDict, total=False):
    """A dictionary containing the optional per-reading metadata of a piece of scan data"""

    #: * -ve: alarm unknown, e.g. device disconnected
    #: * 0: ok, no alarm
    #: * +ve: there is an alarm
    #:
    #: The exact numbers are transport specific
    alarm_severity: int
    #: A descriptive message if there is an alarm
    message: str


T = TypeVar("T")
P = ParamSpec("P")


class Reading(Generic[T], ReadingOptional):
    """A dictionary containing the value and timestamp of a piece of scan data"""

    #: The current value, as a JSON encodable type or numpy array
    value: T
    #: Timestamp in seconds since the UNIX epoch
    timestamp: float


Asset = Union[
    tuple[Literal["resource"], PartialResource],
    tuple[Literal["datum"], Datum],
]


StreamAsset = Union[
    tuple[Literal["stream_resource"], StreamResource],
    tuple[Literal["stream_datum"], StreamDatum],
]


SyncOrAsync = Union[T, Awaitable[T]]
SyncOrAsyncIterator = Union[Iterator[T], AsyncIterator[T]]


@runtime_checkable
class Status(Protocol):
    @abstractmethod
    def add_callback(self, callback: Callable[["Status"], None]) -> None:
        """Add a callback function to be called upon completion.

        The function must take the status as an argument.

        If the Status object is done when the function is added, it should be
        called immediately.
        """
        ...

    @abstractmethod
    def exception(self, timeout: Optional[float] = 0.0) -> Optional[BaseException]: ...

    @property
    @abstractmethod
    def done(self) -> bool:
        """If done return True, otherwise return False."""
        ...

    @property
    @abstractmethod
    def success(self) -> bool:
        """If done return whether the operation was successful."""
        ...


@runtime_checkable
class HasName(Protocol):
    @property
    @abstractmethod
    def name(self) -> str:
        """Used to populate object_keys in the Event DataKey

        https://blueskyproject.io/event-model/event-descriptors.html#object-keys"""
        ...


@runtime_checkable
class HasParent(Protocol):
    @property
    @abstractmethod
    def parent(self) -> Optional[Any]:
        """``None``, or a reference to a parent device.

        Used by the RE to stop duplicate stages.
        """
        ...


@runtime_checkable
class WritesExternalAssets(Protocol):
    @abstractmethod
    def collect_asset_docs(self) -> SyncOrAsyncIterator[Asset]:
        """Create the resource and datum documents describing data in external
            source.

        Example yielded values:

        .. code-block:: python

            ('resource', {
                'path_semantics': 'posix',
                'resource_kwargs': {'frame_per_point': 1},
                'resource_path': 'det.h5',
                'root': '/tmp/tmpcvxbqctr/',
                'spec': 'AD_HDF5',
                'uid': '9123df61-a09f-49ae-9d23-41d4d6c6d788'
            })
            # or
            ('datum', {
                'datum_id': '9123df61-a09f-49ae-9d23-41d4d6c6d788/0',
                'datum_kwargs': {'point_number': 0},
                'resource': '9123df61-a09f-49ae-9d23-41d4d6c6d788'}
            })
        """
        ...


@runtime_checkable
class WritesStreamAssets(Protocol):
    @abstractmethod
    def collect_asset_docs(self, index: Optional[int] = None) -> SyncOrAsyncIterator[StreamAsset]:
        """Create the resource and datum documents describing data in external
            source up to a given index if provided.

            An index will be provided when using stream resources and datums. The asset
            docs will be collected from multiple streams and synchronised on the
            highest common stream index.

        Example yielded values:

        .. code-block:: python

            ('resource', {
                'path_semantics': 'posix',
                'resource_kwargs': {'frame_per_point': 1},
                'resource_path': 'det.h5',
                'root': '/tmp/tmpcvxbqctr/',
                'spec': 'AD_HDF5',
                'uid': '9123df61-a09f-49ae-9d23-41d4d6c6d788'
            })
            # or
            ('datum', {
                'datum_id': '9123df61-a09f-49ae-9d23-41d4d6c6d788/0',
                'datum_kwargs': {'point_number': 0},
                'resource': '9123df61-a09f-49ae-9d23-41d4d6c6d788'}
            })
        """
        ...

    @abstractmethod
    def get_index(self) -> SyncOrAsync[int]:
        """Retrive the current index of writer."""
        ...


@runtime_checkable
class Configurable(Protocol[T]):
    @abstractmethod
    def read_configuration(self) -> SyncOrAsync[dict[str, Reading[T]]]:
        """Same API as ``read`` but for slow-changing fields related to configuration.
        e.g., exposure time. These will typically be read only once per run.

        This can be a standard function or an ``async`` function.
        """
        ...

    @abstractmethod
    def describe_configuration(self) -> SyncOrAsync[dict[str, DataKey]]:
        """Same API as ``describe``, but corresponding to the keys in
        ``read_configuration``.

        This can be a standard function or an ``async`` function.
        """
        ...


@runtime_checkable
class Triggerable(Protocol):
    @abstractmethod
    def trigger(self) -> Status:
        """Return a ``Status`` that is marked done when the device is done triggering."""
        ...


@runtime_checkable
class Preparable(Protocol):
    @abstractmethod
    def prepare(self, value) -> Status:
        """Prepare a device for scanning.

        This method provides similar functionality to ``Stageable.stage`` and
        ``Movable.set``, with key differences:

        ``Stageable.stage``
        ^^^^^^^^^^^^^^^^^^^
        Staging a device translates to, "I'm going to use this in a scan, but I'm
        not sure how". Preparing it translates to, "I'm about to do a step or a fly
        scan with these parameters". Staging should be universal across many different
        types of scans, however prepare is specific to an input value passed in.

        ``Movable.set``
        ^^^^^^^^^^^^^^^
        For some devices, preparation for a scan could involve multiple soft or
        hardware signals being configured and/or set. ``prepare`` therefore allows
        these to be bundled together, along with other logic.

        For example, a Flyable device should have the following methods called on it to
        perform a fly-scan:

            prepare(flyscan_params)
            kickoff()
            complete()

        If the device is a detector, ``collect_asset_docs`` can be called repeatedly
        while ``complete`` is not done to publish frames. Alternatively, to step-scan a
        detector,

            prepare(frame_params) to setup N software triggered frames
            trigger() to take N frames
            collect_asset_docs() to publish N frames

        Returns a Status that is marked done when the device is ready for a scan.
        """
        ...


@runtime_checkable
class Readable(HasName, Protocol[T]):
    @abstractmethod
    def read(self) -> SyncOrAsync[dict[str, Reading[T]]]:
        """Return an OrderedDict mapping string field name(s) to dictionaries
        of values and timestamps and optional per-point metadata.

        This can be a standard function or an ``async`` function.

        Example return value:

        .. code-block:: python

            OrderedDict(('channel1',
                         {'value': 5, 'timestamp': 1472493713.271991}),
                         ('channel2',
                         {'value': 16, 'timestamp': 1472493713.539238}))
        """
        ...

    @abstractmethod
    def describe(self) -> SyncOrAsync[dict[str, DataKey]]:
        """Return an OrderedDict with exactly the same keys as the ``read``
        method, here mapped to per-scan metadata about each field.

        This can be a standard function or an ``async`` function.

        Example return value:

        .. code-block:: python

            OrderedDict(('channel1',
                         {'source': 'XF23-ID:SOME_PV_NAME',
                          'dtype': 'number',
                          'shape': []}),
                        ('channel2',
                         {'source': 'XF23-ID:SOME_PV_NAME',
                          'dtype': 'number',
                          'shape': []}))
        """
        ...


@runtime_checkable
class Collectable(HasName, Protocol):
    @abstractmethod
    def describe_collect(self) -> SyncOrAsync[Union[dict[str, DataKey], dict[str, dict[str, DataKey]]]]:
        """This is like ``describe()`` on readable devices, but with an extra layer of nesting.

        Since a flyer can potentially return more than one event stream, this is either
            * a dict of stream names (strings) mapped to a ``describe()``-type output for each.
            * a ``describe()``-type output of the descriptor name passed in with the ``name``
                argument of the message.

        This can be a standard function or an ``async`` function.
        """
        ...


@runtime_checkable
class EventCollectable(Collectable, Protocol):
    @abstractmethod
    def collect(self) -> SyncOrAsyncIterator[PartialEvent]:
        """Yield dictionaries that are partial Event documents.

        They should contain the keys 'time', 'data', and 'timestamps'.
        A 'uid' is added by the RunEngine.
        """
        ...


@runtime_checkable
class EventPageCollectable(Collectable, Protocol):
    @abstractmethod
    def collect_pages(self) -> SyncOrAsyncIterator[PartialEventPage]:
        """Yield dictionaries that are partial EventPage documents.

        They should contain the keys 'time', 'data', and 'timestamps'.
        A 'uid' is added by the RunEngine.
        """
        ...


T_co = TypeVar("T_co", contravariant=True)


@runtime_checkable
class Movable(Protocol[T_co]):
    @abstractmethod
    def set(self, *args: Unpack[tuple[T_co]]) -> Status:
        """Return a ``Status`` that is marked done when the device is done moving."""
        ...


class Location(Generic[T], TypedDict):
    """A dictionary containing the location of a Device"""

    #: Where the Device was requested to move to
    setpoint: T
    #: Where the Device actually is at the moment
    readback: T


@runtime_checkable
class Locatable(Movable[T], Protocol):
    @abstractmethod
    def locate(self) -> SyncOrAsync[Location[T]]:
        """Return the current location of a Device.

        While a ``Readable`` reports many values, a ``Movable`` will have the
        concept of location. This is where the Device currently is, and where it
        was last requested to move to. This protocol formalizes how to get the
        location from a ``Movable``.
        """
        ...


@runtime_checkable
class Flyable(HasName, Protocol):
    @abstractmethod
    def kickoff(self) -> Status:
        """Begin acculumating data.

        Return a ``Status`` and mark it done when acqusition has begun.
        """
        ...

    @abstractmethod
    def complete(self) -> Status:
        """Return a ``Status`` and mark it done when acquisition has completed."""
        ...


@runtime_checkable
class Stageable(Protocol):
    # TODO: we were going to extend these to be able to return plans, what
    # signature should they have?
    @abstractmethod
    def stage(self) -> Union[Status, list[Any]]:
        """An optional hook for "setting up" the device for acquisition.

        It should return a ``Status`` that is marked done when the device is
        done staging.
        """
        ...

    @abstractmethod
    def unstage(self) -> Union[Status, list[Any]]:
        """A hook for "cleaning up" the device after acquisition.

        It should return a ``Status`` that is marked done when the device is finished
        unstaging.
        """
        ...


@runtime_checkable
class Pausable(Protocol):
    @abstractmethod
    def pause(self) -> SyncOrAsync[None]:
        """Perform device-specific work when the RunEngine pauses.

        This can be a standard function or an ``async`` function.
        """
        ...

    @abstractmethod
    def resume(self) -> SyncOrAsync[None]:
        """Perform device-specific work when the RunEngine resumes after a pause.

        This can be a standard function or an ``async`` function.
        """
        ...


@runtime_checkable
class Stoppable(Protocol):
    @abstractmethod
    def stop(self, success=True) -> SyncOrAsync[None]:
        """Safely stop a device that may or may not be in motion.

        The argument ``success`` is a boolean.
        When ``success`` is true, bluesky is stopping the device as planned
        and the device should stop "normally".
        When ``success`` is false, something has gone wrong and the device
        may wish to take defensive action to make itself safe.

        This can be a standard function or an ``async`` function.
        """
        ...


Callback = Callable[[dict[str, Reading[T]]], None]


@runtime_checkable
class Subscribable(HasName, Protocol[T]):
    @abstractmethod
    def subscribe(self, function: Callback[T]) -> None:
        """Subscribe to updates in value of a device.

        When the device has a new value ready, it should call ``function``
        with something that looks like the output of ``read()``.

        Needed for :doc:`monitored <async>`.
        """
        ...

    @abstractmethod
    def clear_sub(self, function: Callback[T]) -> None:
        """Remove a subscription."""
        ...


@runtime_checkable
class Checkable(Protocol[T_co]):
    @abstractmethod
    def check_value(self, value: T_co) -> SyncOrAsync[None]:
        """Test for a valid setpoint without actually moving.

        This should accept the same arguments as ``set``. It should raise an
        Exception if the argument represent an illegal setting --- e.g. a
        position that would move a motor outside its limits or a temperature
        controller outside of its settable range.

        This method is used by simulators that check limits. If not implemented
        those simulators should assume all values are valid, but may warn.

        This method may be used during a scan, so should not write to any Signals

        This can be a standard function or an ``async`` function.
        """
        ...


class Hints(TypedDict, total=False):
    """A dictionary of optional hints for visualization"""

    #: A list of the interesting fields to plot
    fields: list[str]
    #: Partition fields (and their stream name) into dimensions for plotting
    #:
    #: ``'dimensions': [(fields, stream_name), (fields, stream_name), ...]``
    dimensions: list[tuple[list[str], str]]
    #: Include this if scan data is sampled on a regular rectangular grid
    gridding: Literal["rectilinear", "rectilinear_nonsequential"]


@runtime_checkable
class HasHints(HasName, Protocol):
    @property
    @abstractmethod
    def hints(self) -> Hints:
        """A dictionary of suggestions for best-effort visualization and processing.

        This does not affect what data is read or saved; it is only
        a suggestion to enable automated tools to provide helpful information
        with minimal guidance from the user. See :ref:`hints`.
        """
        ...


@runtime_checkable
class NamedMovable(Movable[T_co], HasHints, Protocol):
    """A movable object that has a name and hints."""

    ...


def check_supports(obj: T, protocol: type[Any]) -> T:
    """Check that an object supports a protocol

    This exists so that multiple protocol checks can be run in a mypy
    compatible way, e.g.::

        triggerable = check_supports(obj, Triggerable)
        triggerable.trigger()
        readable = check_supports(obj, Readable)
        readable.read()
    """
    assert isinstance(obj, protocol), "%s does not implement all %s methods" % (obj, protocol.__name__)  # noqa: UP031
    return obj


# Descriptor with previous name on imports for backwards compatibility.
Descriptor = DataKey



import tempfile

import mypy.api
import pytest
from ophyd import sim

from bluesky import protocols as bs_protocols


def test_status():
    assert isinstance(sim.NullStatus(), bs_protocols.Status)
    assert isinstance(sim.StatusBase(), bs_protocols.Status)


def test_readable():
    assert isinstance(sim.motor1, bs_protocols.Readable)
    assert isinstance(sim.det1, bs_protocols.Readable)
    assert not isinstance(sim.flyer1, bs_protocols.Readable)


def test_movable():
    assert isinstance(sim.motor1, bs_protocols.Movable)
    assert not isinstance(sim.det1, bs_protocols.Movable)
    assert not isinstance(sim.flyer1, bs_protocols.Movable)


def test_flyable():
    assert isinstance(sim.flyer1, bs_protocols.Flyable)
    assert not isinstance(sim.det1, bs_protocols.Flyable)
    assert not isinstance(sim.motor1, bs_protocols.Flyable)


def test_stageable():
    assert isinstance(sim.det1, bs_protocols.Stageable)


def test_pausable():
    assert isinstance(sim.det1, bs_protocols.Pausable)


def test_subscribable():
    assert isinstance(sim.det1, bs_protocols.Subscribable)
    assert isinstance(sim.motor1, bs_protocols.Subscribable)
    assert not isinstance(sim.flyer1, bs_protocols.Subscribable)


def test_checkable():
    assert isinstance(sim.motor1, bs_protocols.Checkable)


# I think the commented out tests pass because __getattr__ is implemented, but not sure
@pytest.mark.skip(reason="ophyd missing py.typed to communicate type hints to mypy")
@pytest.mark.parametrize(
    "type_, hardware, pass_",
    [
        ("Readable", "ABDetector(name='hi')", True),
        ("Readable", "SynAxis(name='motor1')", True),
        ("Readable", "TrivialFlyer()", False),
        ("Configurable", "ABDetector(name='hi')", True),
        # ("Movable", "ABDetector(name='hi')", False),  # mypy passed when it shouldn't
        ("Movable", "SynAxis(name='motor1')", True),
        ("Movable", "TrivialFlyer()", False),
        ("Stageable", "ABDetector(name='hi')", True),
        ("Pausable", "ABDetector(name='hi')", True),
        ("Checkable", "SynAxis(name='motor1')", True),
        # ("Flyable", "ABDetector(name='hi')", False),  # mypy passed when it shouldn't
        # ("Flyable", "SynAxis(name='motor1')", False),  # mypy passed when it shouldn't
        ("Flyable", "TrivialFlyer()", True),
        ("Subscribable", "SynAxis(name='motor1')", True),
    ],
)
def test_mypy(type_, hardware, pass_):
    template = f"""
from bluesky import protocols as bs_protocols
from ophyd import sim

var: bs_protocols.{type_} = sim.{hardware}
"""

    with tempfile.NamedTemporaryFile("wt") as f:
        f.write(template)
        f.seek(0)
        stdout, stderr, exit = mypy.api.run([f.name])
        # pass true means exit 0, pass false means nonzero exit
        try:
            assert exit != pass_
        except AssertionError:
            print(stdout)
            print(stderr)
            raise



import pytest

import bluesky.plans as bp
from bluesky.callbacks.best_effort import BestEffortCallback


def checker(name, doc):
    if name == "start":
        assert "hints" in doc
        assert "dimensions" in doc["hints"]
        assert "uid" in doc


@pytest.mark.parametrize(
    "pln,args,kwargs",
    [
        # Notice that 'args' does not include [det], which is inserted in the body
        # of the test function below.
        (bp.count, (), {}),
        (bp.count, (), {"num": 3}),
        (bp.scan, ("motor", 1, 2, 2), {}),
        (bp.scan, ("motor", 1, 2), {"num": 2}),
        (bp.scan, ("motor1", 1, 2, "motor2", 1, 2, 2), {}),
        (bp.scan, ("motor1", 1, 2, "motor2", 1, 2), {"num": 2}),
        (bp.rel_scan, ("motor", 1, 2, 2), {}),
        (bp.rel_scan, ("motor", 1, 2), {"num": 2}),
        (bp.rel_scan, ("motor1", 1, 2, "motor2", 1, 2, 2), {}),
        (bp.rel_scan, ("motor1", 1, 2, "motor2", 1, 2), {"num": 2}),
        (bp.inner_product_scan, (2, "motor", 1, 2), {}),
        (bp.relative_inner_product_scan, (2, "motor", 1, 2), {}),
        (bp.grid_scan, ("motor1", 1, 2, 2, "motor2", 1, 2, 3, False), {}),
        (bp.spiral, ("motor1", "motor2", 0.0, 0.0, 0.3, 0.3, 0.05, 3), {}),
        (bp.rel_spiral, ("motor1", "motor2", 0.3, 0.3, 0.05, 3), {}),
        (bp.spiral_fermat, ("motor1", "motor2", 0.0, 0.0, 0.3, 0.3, 0.05, 3), {}),
        (bp.rel_spiral_fermat, ("motor1", "motor2", 0.3, 0.3, 0.05, 3), {}),
    ],
)
def test_plans(RE, pln, args, kwargs, hw):
    args = tuple(getattr(hw, v, v) if isinstance(v, str) else v for v in args)
    det = hw.det
    bec = BestEffortCallback()
    RE.subscribe(bec)
    RE.subscribe(checker)
    RE(pln([det], *args, **kwargs))


@pytest.mark.parametrize(
    "pln,args,kwargs",
    [
        # repeat with motor objects that do not have hints
        (bp.scan, ("motor_no_hints1", 1, 2, 2), {}),
        (bp.inner_product_scan, (2, "motor_no_hints1", 1, 2), {}),
        (bp.relative_inner_product_scan, (2, "motor_no_hints1", 1, 2), {}),
        (bp.grid_scan, ("motor_no_hints1", 1, 2, 2, "motor_no_hints2", 1, 2, 3, False), {}),
        (bp.spiral, ("motor_no_hints1", "motor_no_hints2", 0.0, 0.0, 0.3, 0.3, 0.05, 3), {}),
        (bp.rel_spiral, ("motor_no_hints1", "motor_no_hints2", 0.3, 0.3, 0.05, 3), {}),
        (bp.spiral_fermat, ("motor_no_hints1", "motor_no_hints2", 0.0, 0.0, 0.3, 0.3, 0.05, 3), {}),
        (bp.rel_spiral_fermat, ("motor_no_hints1", "motor_no_hints2", 0.3, 0.3, 0.05, 3), {}),
    ],
)
@pytest.mark.xfail
def test_plans_motors_no_hints(RE, pln, args, kwargs, hw):
    args = tuple(getattr(hw, v, v) if isinstance(v, str) else v for v in args)
    det = hw.det
    for v in args:
        assert not hasattr(v, "hints")
    bec = BestEffortCallback()
    RE.subscribe(bec)
    RE(pln([det], *args, **kwargs))


@pytest.mark.parametrize(
    "pln,args,kwargs",
    [
        # repeat with motor objects that have empty hints
        (bp.scan, ("motor1", 1, 2, 2), {}),
        (bp.inner_product_scan, (2, "motor1", 1, 2), {}),
        (bp.relative_inner_product_scan, (2, "motor1", 1, 2), {}),
        (bp.grid_scan, ("motor1", 1, 2, 2, "motor2", 1, 2, 3, False), {}),
        (bp.spiral, ("motor1", "motor2", 0.0, 0.0, 0.3, 0.3, 0.05, 3), {}),
        (bp.rel_spiral, ("motor1", "motor2", 0.3, 0.3, 0.05, 3), {}),
        (bp.spiral_fermat, ("motor1", "motor2", 0.0, 0.0, 0.3, 0.3, 0.05, 3), {}),
        (bp.rel_spiral_fermat, ("motor1", "motor2", 0.3, 0.3, 0.05, 3), {}),
    ],
)
@pytest.mark.xfail
def test_plans_motor_empty_hints(RE, pln, args, kwargs, hw):
    args = tuple(getattr(hw, v, v) if isinstance(v, str) else v for v in args)
    for v in args:
        if hasattr(v, "hints"):
            v.hints = {}
            assert v.hints == {}
    det = hw.det
    bec = BestEffortCallback()
    RE.subscribe(bec)
    RE(pln([det], *args, **kwargs))



import collections.abc
import gc

import numpy
import pytest
from numpy.testing import assert_array_equal
from packaging.version import Version

from ..plans import count
from ..utils import PersistentDict

zict = pytest.importorskip("zict")
pytestmark = pytest.mark.skipif(
    condition=Version(zict.__version__) >= Version("3"),
    reason="Version 3 does not support multiple instances looking at same files",
)


def test_persistent_dict(tmp_path):
    d = PersistentDict(tmp_path)
    d["a"] = 1
    d["b"] = (1, 2)
    d["c"] = numpy.zeros((5, 5))
    d["d"] = {"a": 10, "b": numpy.ones((5, 5))}
    expected = dict(d)
    actual = PersistentDict(tmp_path)
    recursive_assert_equal(actual, expected)

    # Update a value and check again.
    d["a"] = 2
    expected = dict(d)
    actual.reload()  # Force load changes from disk
    recursive_assert_equal(actual, expected)

    # Test element deletion
    del d["b"]
    with pytest.raises(KeyError):
        d["b"]
    assert "b" not in d

    # Smoke test the accessor and the __repr__.
    assert d.directory == tmp_path
    d.__repr__()


def test_persistent_dict_mutable_value(tmp_path):
    d = PersistentDict(tmp_path)
    d["a"] = []
    d["a"].append(1)
    expected = {"a": [1]}
    # Check the in-memory version is updated
    recursive_assert_equal(d, expected)
    # Check that the __repr__ reflects this.
    assert "{'a': [1]}" in repr(d)
    # Check that contents are synced to disk at exit.
    del d
    gc.collect()
    actual = PersistentDict(tmp_path)
    recursive_assert_equal(actual, expected)


def test_pop(tmp_path):
    d = PersistentDict(tmp_path)
    d["a"] = 1
    d["b"] = 2
    d.pop("b")
    expected = {"a": 1}
    # Check the in-memory version is updated
    recursive_assert_equal(d, expected)
    # Check that the __repr__ reflects this.
    assert "{'a': 1}" in repr(d)
    # Check that contents are synced to disk at exit.
    del d
    gc.collect()
    actual = PersistentDict(tmp_path)
    recursive_assert_equal(actual, expected)


def test_popitem(tmp_path):
    d = PersistentDict(tmp_path)
    d["a"] = 1
    d["b"] = 2
    d.popitem()
    expected = {"a": 1}
    # Check the in-memory version is updated
    recursive_assert_equal(d, expected)
    # Check that the __repr__ reflects this.
    assert "{'a': 1}" in repr(d)
    # Check that contents are synced to disk at exit.
    del d
    gc.collect()
    actual = PersistentDict(tmp_path)
    recursive_assert_equal(actual, expected)


def test_update(tmp_path):
    d = PersistentDict(tmp_path)
    d.update(a=1)
    expected = {"a": 1}
    # Check the in-memory version is updated
    recursive_assert_equal(d, expected)
    # Check that the __repr__ reflects this.
    assert "{'a': 1}" in repr(d)
    # Check that contents are synced to disk at exit.
    del d
    gc.collect()
    actual = PersistentDict(tmp_path)
    recursive_assert_equal(actual, expected)


def test_setdefault(tmp_path):
    d = PersistentDict(tmp_path)
    d.setdefault("a", 1)
    expected = {"a": 1}
    # Check the in-memory version is updated
    recursive_assert_equal(d, expected)
    # Check that the __repr__ reflects this.
    assert "{'a': 1}" in repr(d)
    # Check that contents are synced to disk at exit.
    del d
    gc.collect()
    actual = PersistentDict(tmp_path)
    recursive_assert_equal(actual, expected)


def test_clear(tmp_path):
    d = PersistentDict(tmp_path)
    d["a"] = 1
    d.clear()
    expected = {}
    # Check the in-memory version is updated
    recursive_assert_equal(d, expected)
    # Check that the __repr__ reflects this.
    assert "{}" in repr(d)
    # Check that contents are synced to disk at exit.
    del d
    gc.collect()
    actual = PersistentDict(tmp_path)
    recursive_assert_equal(actual, expected)


def test_integration(tmp_path, RE, hw):
    """
    Test integration with RE.

    Not looking for anything *specific* here, just general paranoia in case
    unforseen future changes create a bad interaction between PersistentDict
    and RE, as happened with HistoryDict and RE.
    """
    d = PersistentDict(tmp_path)
    d["a"] = 1
    d["b"] = (1, 2)
    d["c"] = numpy.zeros((5, 5))
    d["d"] = {"a": 10, "b": numpy.ones((5, 5))}
    expected = dict(d)
    expected["scan_id"] = 1

    RE.md = d
    RE(count([hw.det]))
    recursive_assert_equal(RE.md, expected)

    reloaded = PersistentDict(tmp_path)
    recursive_assert_equal(reloaded, expected)


def recursive_assert_equal(actual, expected):
    assert set(actual.keys()) == set(expected.keys())
    for key in actual:
        if isinstance(actual[key], collections.abc.MutableMapping):
            recursive_assert_equal(actual[key], expected[key])
        else:
            assert_array_equal(actual[key], expected[key])



from ._version import __version__
from .log import set_handler
from .preprocessors import SupplementalData
from .run_engine import RunEngine
from .utils import FailedStatus, IllegalMessageSequence, Msg, RunEngineInterrupted

__all__ = [
    "__version__",
    "set_handler",
    "SupplementalData",
    "RunEngine",
    "FailedStatus",
    "IllegalMessageSequence",
    "Msg",
    "RunEngineInterrupted",
]



import argparse
import logging
import threading

from bluesky.callbacks.zmq import Proxy, RemoteDispatcher

logger = logging.getLogger("bluesky")


def start_dispatcher(host, port, logfile=None):
    """The dispatcher function
    Parameters
    ----------
    logfile : string
        string come from user command. ex --logfile=temp.log
        logfile will be "temp.log". logfile could be empty.
    """
    dispatcher = RemoteDispatcher((host, port))
    if logfile is not None:
        raise ValueError(
            "Parameter 'logfile' is deprecated and will be removed in future releases. "
            "Currently it does not have effect. Call the function with 'logfile=None' "
        )

    def log_writer(name, doc):
        """logger's wrapper function
        This function will be used to fit .subscribe() method.
        It has two arguments as .subscribe expects. Inside, it
        calls logger.* to write doc which is a dict as a str
        into logfile
        """
        if name in ("start", "stop"):
            logger.info("%s: %r", name, doc)
        else:
            logger.debug("%s: %r", name, doc)

    dispatcher.subscribe(log_writer)  # Subscribe log writer
    dispatcher.start()


def main():
    DESC = "Start a 0MQ proxy for publishing bluesky documents over a network."
    parser = argparse.ArgumentParser(description=DESC)
    parser.add_argument("in_port", type=int, nargs=1, help="port that RunEngines should broadcast to")
    parser.add_argument("out_port", type=int, nargs=1, help="port that subscribers should subscribe to")
    parser.add_argument(
        "--verbose",
        "-v",
        action="count",
        help=("Show 'start' and 'stop' documents. (Use -vvv to show all documents.)"),
    )
    parser.add_argument("--logfile", type=str, help="Redirect logging output to a file on disk.")
    args = parser.parse_args()
    in_port = args.in_port[0]
    out_port = args.out_port[0]

    if args.verbose:
        from bluesky.log import config_bluesky_logging

        # "INFO" if called with '-v' or '-vv', "DEBUG" if called with '-vvv'
        level = "INFO" if args.verbose <= 2 else "DEBUG"
        if args.logfile:
            config_bluesky_logging(level=level, file=args.logfile)
        else:
            config_bluesky_logging(level=level)
        # Set daemon to kill all threads upon IPython exit
        threading.Thread(target=start_dispatcher, args=("localhost", out_port), daemon=True).start()

    print("Connecting...")
    proxy = Proxy(in_port, out_port)
    print("Receiving on port %d; publishing to port %d." % (in_port, out_port))
    print("Use Ctrl+C to exit.")
    try:
        proxy.start()
    except KeyboardInterrupt:
        print("Interrupted. Exiting...")


if __name__ == "__main__":
    main()



import importlib.metadata
from collections.abc import Iterable, Iterator
from random import randint, sample
from typing import Literal, Optional, Union

import packaging.version
import pytest
from event_model import ComposeStreamResource, EventModelValueError
from event_model.documents.event_descriptor import DataKey
from event_model.documents.event_page import PartialEventPage
from event_model.documents.stream_datum import StreamRange
from event_model.documents.stream_resource import StreamResource

from bluesky import Msg
from bluesky.protocols import (
    EventPageCollectable,
    Flyable,
    Pausable,
    Readable,
    Reading,
    StreamAsset,
    SyncOrAsync,
    WritesExternalAssets,
)
from bluesky.run_engine import RunEngineInterrupted

event_model_version = packaging.version.parse(importlib.metadata.version("event_model"))
DRAFT_0_STREAM_RESOURCE = event_model_version < packaging.version.parse("1.21.0")


class ExternalAssetDevice:
    sequence_counter_at_chunks: Optional[Union[range, list[int]]] = None
    current_chunk: int = 0

    def __init__(
        self,
        number_of_chunks: int,
        number_of_frames: int,
        detectors: Optional[list[str]] = None,
        stream_datum_contains_one_index: bool = False,
    ):
        self.detectors = detectors or ["det1", "det2", "det3"]
        self.compose_stream_resource = ComposeStreamResource()
        if DRAFT_0_STREAM_RESOURCE:
            self.stream_resource_compose_datum_pairs = tuple(
                self.compose_stream_resource("", "", f"non_existent_{det}.hdf5", det, {})  # type: ignore[arg-type]
                for det in self.detectors
            )
        else:
            self.stream_resource_compose_datum_pairs = tuple(
                self.compose_stream_resource(
                    mimetype="application/x-hdf5",
                    uri=f"file://localhost/non_existent_{det}.hdf5",
                    data_key=det,
                    parameters={"path": "/data"},
                )
                for det in self.detectors
            )
        # Number of collect calls that will be made
        self.number_of_chunks = number_of_chunks

        # Number of frames in a detector in the run
        self.number_of_frames = number_of_frames

        if not stream_datum_contains_one_index:
            # To simulate jitter
            self.sequence_counter_at_chunks = sorted(
                sample(range(1, self.number_of_frames), self.number_of_chunks - 1)
            )

        else:
            # If there's no jitter and every index is {start:n, stop:n+1}
            self.sequence_counter_at_chunks = range(1, self.number_of_chunks)

        self.new_random_width = self.sequence_counter_at_chunks[0]

    def collect_resources(self) -> Iterator[tuple[Literal["stream_resource"], StreamResource]]:
        for stream_resource, _ in self.stream_resource_compose_datum_pairs:
            yield ("stream_resource", stream_resource)

    def collect_stream_datum(self) -> Iterator[StreamAsset]:
        assert self.sequence_counter_at_chunks is not None
        if self.current_chunk >= self.number_of_chunks:  # No more collect()
            return

        elif self.current_chunk == 0:  # First collect()
            yield from self.collect_resources()

        elif self.current_chunk == self.number_of_chunks - 1:  # Last collect()
            self.new_random_width = (
                self.number_of_frames - self.sequence_counter_at_chunks[self.number_of_chunks - 2]
            )
        else:
            self.new_random_width = (
                self.sequence_counter_at_chunks[self.current_chunk]
                - self.sequence_counter_at_chunks[self.current_chunk - 1]
            )

        self.current_chunk += 1

        for _, compose_stream_datum in self.stream_resource_compose_datum_pairs:
            indices_start = randint(0, 250)
            yield (
                "stream_datum",
                compose_stream_datum(
                    indices=StreamRange(start=indices_start, stop=indices_start + self.new_random_width)
                ),
            )

    def collect_stream_datum_mismatched_indices(self):
        if self.current_chunk < self.number_of_chunks:
            yield from self.collect_stream_datum()

        indices_start = randint(0, 250)
        yield from (
            (
                "stream_datum",
                compose_stream_datum(indices=StreamRange(start=indices_start, stop=indices_start + i)),
            )
            for i, (_, compose_stream_datum) in enumerate(self.stream_resource_compose_datum_pairs)
        )

    def collect_stream_datum_repeat_stream_resource(self) -> Iterator[StreamAsset]:
        if self.current_chunk == int(self.number_of_chunks / 2):
            # New stream_resource half way through the run
            if DRAFT_0_STREAM_RESOURCE:
                self.stream_resource_compose_datum_pairs = tuple(
                    self.compose_stream_resource("", "", f"non_existent_{det}.hdf5", det, {})  # type: ignore[arg-type]
                    for det in self.detectors
                )
            else:
                self.stream_resource_compose_datum_pairs = tuple(
                    self.compose_stream_resource(
                        mimetype="application/x-hdf5",
                        uri=f"file://localhost/non_existent_{det}.hdf5",
                        data_key=det,
                        parameters={"path": "/data"},
                    )
                    for det in self.detectors
                )
            yield from self.collect_resources()

        yield from self.collect_stream_datum()

    def collect_Pageable_with_name(self) -> SyncOrAsync[Iterable[PartialEventPage]]:
        def timestamps():
            return {"det4": [1] * self.new_random_width}

        def data():
            return {"det4": ["a"] * self.new_random_width}

        return [PartialEventPage(timestamps=timestamps(), data=data(), time=[1] * self.new_random_width)]


def kickoff(self, *_, **__):
    class X:
        def add_callback(cls, *_, **__): ...

    return X


def complete(self, *_, **__):
    class X:
        def add_callback(cls, *_, **__): ...

    return X


def pause(self, *_, **__): ...


def resume(self, *_, **__): ...


def read_Readable(self) -> dict[str, Reading]:
    return dict(det3=dict(value=1.2, timestamp=0.0))  # noqa: C408


def describe_Readable(self) -> dict[str, DataKey]:
    return dict(  # noqa: C408
        det1=dict(source="hw1", dtype="number", shape=[], external="STREAM:"),  # noqa: C408
        det2=dict(source="hw1", dtype="number", shape=[], external="STREAM:"),  # noqa: C408
        det3=dict(source="hw2", dtype="number", shape=[]),  # noqa: C408
    )


def describe_collect_with_name(self) -> SyncOrAsync[dict[str, DataKey]]:
    description = {
        str(det): DataKey(shape=[], source="stream1", dtype="string", external="STREAM:")
        for det in ["det1", "det2", "det3"]
    }
    description.update({"det4": DataKey(shape=[], source="stream1", dtype="string")})
    return description


def test_flyscan_with_stream_datum_pause(RE):
    external_asset_device = ExternalAssetDevice(10, 100)

    class DummyDeviceFlyableEventPageCollectablePausable(Flyable, EventPageCollectable, Pausable):
        kickoff = kickoff
        complete = complete
        pause = pause
        resume = resume
        collect_pages = external_asset_device.collect_Pageable_with_name
        collect_asset_docs = external_asset_device.collect_stream_datum
        describe_collect = describe_collect_with_name
        name = "x"

    x = DummyDeviceFlyableEventPageCollectablePausable()
    collector = []

    with pytest.raises(RunEngineInterrupted):
        RE(
            [
                Msg("open_run", x),
                Msg("kickoff", x),
                Msg("pause", x),
                Msg("declare_stream", None, x, name="primary", collect=True),
                *[Msg("collect", x, name="primary")] * 10,
                Msg("complete", x),
                Msg("close_run", x),
            ],
            lambda *args: collector.append(args),
        )

    RE.resume()

    stream_datum_collected_docs = [doc[1] for doc in collector if doc[0] == "stream_datum"]
    assert len(stream_datum_collected_docs) == 30

    collector_iter = iter(collector)
    stream_resource_collected = []
    # Check for contiguous stream_resource
    try:
        while collector_iter:
            next_collected_doc = next(collector_iter)
            if next_collected_doc[0] == "stream_resource":
                for _ in range(3):
                    assert next_collected_doc[0] == "stream_resource"
                    stream_resource_collected.append(next_collected_doc)
                    next_collected_doc = next(collector_iter)
    except StopIteration:
        ...

    assert stream_resource_collected
    assert len(stream_resource_collected) == 3

    # There are 30 stream datum all together, every 3 will have the same seq_nums
    for i in range(0, 30, 3):
        stream_datum_chunk = stream_datum_collected_docs[i : i + 3]
        seq_nums = stream_datum_chunk[0]["seq_nums"]
        for stream_datum in stream_datum_chunk[-2:]:
            assert stream_datum["seq_nums"] == seq_nums

    assert stream_datum_collected_docs[-1]["seq_nums"]["stop"] == 101


def test_flyscan_with_mismatched_indices(RE):
    external_asset_device = ExternalAssetDevice(9, 100)

    class DummyDeviceFlyableEventPageCollectablePausableMismatchedStreamDatumIndices(
        Flyable, EventPageCollectable, Pausable
    ):
        kickoff = kickoff
        complete = complete
        pause = pause
        resume = resume
        collect_pages = external_asset_device.collect_Pageable_with_name
        collect_asset_docs = external_asset_device.collect_stream_datum_mismatched_indices
        describe_collect = describe_collect_with_name
        name = "x"

    x = DummyDeviceFlyableEventPageCollectablePausableMismatchedStreamDatumIndices()
    collector = []
    plan = [
        Msg("open_run", x),
        Msg("kickoff", x),
        Msg("pause", x),
        Msg("declare_stream", None, x, name="stream1", collect=True),
        *[Msg("collect", x, name="stream1")] * 10,
        Msg("complete", x),
        Msg("close_run", x),
    ]

    with pytest.raises(RunEngineInterrupted):
        RE(plan, lambda *args: collector.append(args))

    with pytest.raises(EventModelValueError):
        RE.resume()


def test_rd_desc_with_declare_stream(RE):
    external_asset_device = ExternalAssetDevice(
        10, 10, detectors=["det1", "det2"], stream_datum_contains_one_index=True
    )

    class DummyDeviceReadablePausableExternalAssets(Readable, Pausable, WritesExternalAssets):
        read = read_Readable
        describe = describe_Readable
        pause = pause
        resume = resume
        collect_asset_docs = external_asset_device.collect_stream_datum
        name = "x"

    x = DummyDeviceReadablePausableExternalAssets()
    collector = []
    plan = [
        Msg("open_run", x),
        *[
            Msg("create", name="hw1"),
            Msg("read", x),
            Msg("save", x),
        ]
        * 10,
        Msg("close_run", x),
    ]

    RE(plan, lambda *args: collector.append(args))

    stream_datums_in_collector = [doc[1] for doc in collector if doc[0] == "stream_datum"]
    assert len(stream_datums_in_collector) == 20
    assert stream_datums_in_collector[-1]["seq_nums"]["stop"] == 11


def test_changing_stream_resource_after_stream_datum_emitted(RE):
    external_asset_device = ExternalAssetDevice(10, 100)

    class DummyDeviceFlyableEventPageCollectable(Flyable, EventPageCollectable):
        kickoff = kickoff
        complete = complete
        collect_pages = external_asset_device.collect_Pageable_with_name
        collect_asset_docs = external_asset_device.collect_stream_datum_repeat_stream_resource
        describe_collect = describe_collect_with_name
        name = "x"

    x = DummyDeviceFlyableEventPageCollectable()
    collector = []
    plan = [
        Msg("open_run", x),
        Msg("kickoff", x),
        Msg("declare_stream", None, x, name="primary", collect=True),
        *[Msg("collect", x, name="primary")] * 10,
        Msg("complete", x),
        Msg("close_run", x),
    ]

    RE(plan, lambda *args: collector.append(args))

    stream_datum_collected_docs = [doc[1] for doc in collector if doc[0] == "stream_datum"]
    assert len(stream_datum_collected_docs) == 30

    collector_iter = iter(collector)
    stream_resource_collected = []
    # Check for contiguous stream_resource
    try:
        while collector_iter:
            next_collected_doc = next(collector_iter)
            if next_collected_doc[0] == "stream_resource":
                for _ in range(3):
                    assert next_collected_doc[0] == "stream_resource"
                    stream_resource_collected.append(next_collected_doc)
                    next_collected_doc = next(collector_iter)
    except StopIteration:
        ...

    assert stream_resource_collected
    assert len(stream_resource_collected) == 6

    # There are 30 stream datum all together, every 3 will have the same seq_nums
    for i in range(0, 30, 3):
        stream_datum_chunk = stream_datum_collected_docs[i : i + 3]
        seq_nums = stream_datum_chunk[0]["seq_nums"]
        for stream_datum in stream_datum_chunk[-2:]:
            assert stream_datum["seq_nums"] == seq_nums

    assert stream_datum_collected_docs[-1]["seq_nums"]["stop"] == 101



import os
from types import ModuleType
from typing import Optional

import pytest

# some module level globals.
ophyd: Optional[ModuleType]
ophyd = None
reason = ""

try:
    import ophyd  # type: ignore
except ImportError as ie:
    # pytestmark = pytest.mark.skip
    ophyd = None
    reason = str(ie)

# define a skip condition based on if ophyd is available or not
requires_ophyd = pytest.mark.skipif(ophyd is None, reason=reason)

uses_os_kill_sigint = pytest.mark.skipif(
    os.name == "nt", reason="os.kill on windows ignores signal argument and kills the entire process."
)



import asyncio
import sys

import pytest

import bluesky.plan_stubs as bps
from bluesky.run_engine import autoawait_in_bluesky_event_loop, call_in_bluesky_event_loop, in_bluesky_event_loop


def test_in_bluesky_event_loop(RE):
    def plan():
        yield from bps.null()
        assert in_bluesky_event_loop()

    assert not in_bluesky_event_loop()
    RE(plan())
    assert not in_bluesky_event_loop()


def test_call_in_bluesky_event_loop(RE):
    event_loop = None

    async def check():
        nonlocal event_loop
        event_loop = asyncio.get_running_loop()

    if sys.version_info >= (3, 10):
        # For some reason asyncio.run reuses the RE loop, then closes
        # it at the end on python 3.9 and below, which makes test_examples.py
        # fail, so skip this bit for that python
        asyncio.run(check())
        assert event_loop and event_loop != RE._loop

    call_in_bluesky_event_loop(check())
    assert event_loop == RE._loop


def test_autoawait_in_bluesky_event_loop(RE):
    event_loop = None

    async def check():
        nonlocal event_loop
        event_loop = asyncio.get_running_loop()

    IPython = pytest.importorskip("IPython")
    ip = IPython.core.interactiveshell.InteractiveShell(user_ns=locals())
    # Check that without the autoawait we get the wrong event loop
    ip.run_cell("await check()")
    assert event_loop and event_loop != RE._loop
    # Then with it we get the right one
    autoawait_in_bluesky_event_loop(ip)
    ip.run_cell("await check()")
    assert event_loop == RE._loop



# %matplotlib qt5
import shutil
from collections.abc import Mapping
from typing import Any

import matplotlib.pyplot as plt
import numpy as np
from databroker import Broker
from databroker.v0 import temp_config
from ophyd.sim import det4, motor1, motor2

from bluesky import RunEngine
from bluesky.callbacks.best_effort import BestEffortCallback
from bluesky.plan_stubs import mov
from bluesky.plans import (
    inner_product_scan,
    outer_product_scan,
    relative_inner_product_scan,
    relative_outer_product_scan,
    scan_nd,
)

plt.ion()

# db setup
config = temp_config()
tempdir = config["metadatastore"]["config"]["directory"]


def cleanup():
    shutil.rmtree(tempdir)


db = Broker.named("temp")

RE = RunEngine({})
# subscribe BEC
bec = BestEffortCallback()
RE.subscribe(bec)
RE.subscribe(db.insert)

# move motor to a reproducible location
RE(mov(motor1, 0))  # type: ignore
RE(mov(motor2, 0))  # type: ignore
RE(relative_outer_product_scan([det4], motor1, -1, 0, 10, motor2, -2, 0, 20, True))
RE(outer_product_scan([det4], motor1, -1, 0, 10, motor2, -2, 0, 20, True))

# move motor to a reproducible location
RE(mov(motor1, 0))  # type: ignore
RE(mov(motor2, 0))  # type: ignore
RE(relative_inner_product_scan([det4], 10, motor1, -1, 0, motor2, -2, 0))
RE(inner_product_scan([det4], 10, motor1, -1, 0, motor2, -2, 0))

# do it manually
from cycler import cycler  # noqa: E402

mot1_cycl = cycler(motor1, np.linspace(-1, 0, 10))
mot2_cycl = cycler(motor2, np.linspace(-2, 0, 10))


# inner product
inner_hints = {"fields": [det4.name], "dimensions": [([motor1.name, motor2.name], "primary")]}
RE(scan_nd([det4], mot1_cycl + mot2_cycl), hints=inner_hints)

# outer product
outer_hints: Mapping[str, Any] = {
    "fields": [det4.name],
    "dimensions": [([motor2.name], "primary"), ([motor1.name], "primary")],
}
md = {
    "shape": (20, 20),
    # 'extents': ([-1, 0], [-1, 0]),
    "hints": outer_hints,
}

RE(scan_nd([det4], mot1_cycl * mot2_cycl), **md)

# make 40 points just to test
mot2_cycl = cycler(motor2, np.linspace(-1, 0, 20))
# now try rectilinear gridding
outer_hints = {
    "fields": [det4.name],
    "dimensions": [([motor2.name], "primary"), ([motor1.name], "primary")],
    "gridding": "rectilinear",
}
md = {
    "shape": (10, 20),
    "extents": ([-1, 0], [-1, 0]),
    "hints": outer_hints,
}

RE(scan_nd([det4], mot1_cycl * mot2_cycl), **md)



from unittest.mock import MagicMock

import pytest

import bluesky.plan_stubs as bps
from bluesky.preprocessors import (
    contingency_decorator,
    contingency_wrapper,
    msg_mutator,
)
from bluesky.run_engine import RequestStop, RunEngine


def test_given_a_plan_that_raises_contigency_will_call_except_plan_with_exception_and_run_engine_errors():
    expected_exception = Exception()

    def except_plan(exception: Exception):
        assert exception == expected_exception
        yield from bps.null()

    # Mock so we can assert called
    except_plan = MagicMock(side_effect=except_plan)

    @contingency_decorator(except_plan=except_plan)
    def raising_plan():
        yield from bps.null()
        raise expected_exception

    RE = RunEngine()

    with pytest.raises(Exception) as exception:
        RE(raising_plan())
        assert exception == expected_exception

    except_plan.assert_called_once()


def test_given_a_plan_that_raises_contigency_with_no_auto_raise_will_call_except_plan_and_RE_does_not_raise():
    expected_exception = Exception()
    expected_return_value = "test"

    def except_plan(exception: Exception):
        assert exception == expected_exception
        yield from bps.null()
        return expected_return_value

    # Mock so we can assert called
    except_plan = MagicMock(side_effect=except_plan)

    @contingency_decorator(except_plan=except_plan, auto_raise=False)
    def raising_plan():
        yield from bps.null()
        raise expected_exception

    RE = RunEngine(call_returns_result=True)

    returned_value = RE(raising_plan())

    except_plan.assert_called_once()
    assert returned_value.plan_result == expected_return_value


def test_given_a_plan_that_raises_contigency_with_no_auto_raise_and_except_plan_that_reraises_run_engine_errors():
    expected_exception = Exception()

    def except_plan(exception: Exception):
        assert exception == expected_exception
        yield from bps.null()
        raise exception

    # Mock so we can assert called
    except_plan = MagicMock(side_effect=except_plan)

    @contingency_decorator(except_plan=except_plan, auto_raise=False)
    def raising_plan():
        yield from bps.null()
        raise expected_exception

    RE = RunEngine()

    with pytest.raises(Exception) as exception:
        RE(raising_plan())
        assert exception == expected_exception

    except_plan.assert_called_once()


def test_exceptions_through_msg_mutator():
    from bluesky import Msg

    def outer():
        for j in range(50):
            yield Msg(f"step {j}")

    def attach(msg):
        cmd = msg.command
        return msg._replace(command=f"{cmd}+")

    def except_plan(e):
        yield Msg("handle it")

    gen = msg_mutator(contingency_wrapper(outer(), except_plan=except_plan), attach)

    msgs = []

    msgs.append(next(gen))
    msgs.append(next(gen))
    msgs.append(next(gen))
    msgs.append(next(gen))
    msgs.append(gen.throw(RequestStop))
    try:
        while True:
            msgs.append(next(gen))
    except RequestStop:
        pass
    else:
        raise False  # noqa: B016
    assert ["step 0+", "step 1+", "step 2+", "step 3+", "handle it+"] == [m.command for m in msgs]



import asyncio
import inspect
import time as ttime
from collections import defaultdict, deque
from collections.abc import Iterable
from itertools import combinations
from logging import LoggerAdapter
from typing import Any, Callable, Literal, Optional, Union, cast

from event_model import (
    ComposeDescriptorBundle,
    DataKey,
    Datum,
    DocumentNames,
    EventModelValueError,
    Resource,
    StreamDatum,
    StreamRange,
    StreamResource,
    compose_run,
    pack_event_page,
)
from event_model.documents.event import Event

from .log import doc_logger
from .protocols import (
    Asset,
    Callback,
    Collectable,
    Configurable,
    EventCollectable,
    EventPageCollectable,
    Flyable,
    HasName,
    Readable,
    Reading,
    StreamAsset,
    Subscribable,
    SyncOrAsync,
    T,
    WritesStreamAssets,
    check_supports,
)
from .utils import (
    IllegalMessageSequence,
    Msg,
    _rearrange_into_parallel_dicts,
    iterate_maybe_async,
    maybe_await,
    maybe_collect_asset_docs,
    maybe_update_hints,
    new_uid,
    short_uid,
)

ObjDict = dict[Any, dict[str, T]]
ExternalAssetDoc = Union[Datum, Resource, StreamDatum, StreamResource]


def _describe_collect_dict_is_valid(
    describe_collect_dict: Union[Any, dict[str, Any]],
) -> bool:  # TODO: change to TypeGuard[dict[str, DataKey]] after python 3.9
    """
    Check if the describe_collect dictionary contains valid DataKeys.
    """
    if not isinstance(describe_collect_dict, dict):
        return False
    return all(
        isinstance(value, dict) and isinstance(value.get("source", None), str)
        for value in describe_collect_dict.values()
    )


class RunBundler:
    def __init__(
        self,
        md: Optional[dict],
        record_interruptions: bool,
        emit: Callable,
        emit_sync: Callable,
        log: LoggerAdapter,
        *,
        strict_pre_declare: bool,
    ):
        # if create can YOLO implicitly create a stream
        self._strict_pre_declare = strict_pre_declare
        # state stolen from the RE
        self.bundling = False  # if we are in the middle of bundling readings
        self._bundle_name = None  # name given to event descriptor
        self._run_start_uid = None  # The (future) runstart uid
        self._objs_read: deque[HasName] = deque()  # objects read in one Event
        self._read_cache: deque[dict[str, Reading]] = deque()  # cache of obj.read() in one Event
        self._asset_docs_cache: deque[Union[Asset, StreamAsset]] = deque()  # cache of obj.collect_asset_docs()
        self._describe_cache: ObjDict[DataKey] = dict()  # cache of all obj.describe() output  # noqa: C408
        self._describe_collect_cache: dict[Any, Union[dict[str, DataKey], dict[str, dict[str, DataKey]]]] = dict()  # noqa: C408  # cache of all obj.describe() output

        self._config_desc_cache: ObjDict[DataKey] = dict()  # " obj.describe_configuration()  # noqa: C408
        self._config_values_cache: ObjDict[Any] = dict()  # " obj.read_configuration() values  # noqa: C408
        self._config_ts_cache: ObjDict[Any] = dict()  # " obj.read_configuration() timestamps  # noqa: C408
        # cache of {name: (doc, compose_event, compose_event_page)}
        self._descriptors: dict[Any, ComposeDescriptorBundle] = dict()  # noqa: C408
        self._descriptor_objs: dict[str, dict[HasName, dict[str, DataKey]]] = dict()  # noqa: C408
        # cache of {obj: {objs_frozen_set: (doc, compose_event, compose_event_page)}
        self._local_descriptors: dict[Any, dict[frozenset[str], ComposeDescriptorBundle]] = dict()  # noqa: C408
        # a seq_num counter per stream
        self._sequence_counters: dict[Any, int] = dict()  # noqa: C408
        self._sequence_counters_copy: dict[Any, int] = dict()  # for if we redo data-points  # noqa: C408
        self._monitor_params: dict[Subscribable, tuple[Callback, dict]] = dict()  # noqa: C408  # cache of {obj: (cb, kwargs)}
        # a cache of stream_resource uid to the data_keys that stream_resource collects for
        self._stream_resource_data_keys: dict[str, Iterable[str]] = dict()  # noqa: C408
        self.run_is_open = False
        self._uncollected: set[HasName] = set()  # objects after kickoff(), before collect()
        # we expect the RE to take care of the composition
        self._md = md
        # this is state on the RE, mirror it here rather than refer to
        # the parent
        self.record_interruptions = record_interruptions
        # this is RE.emit, but lifted to this context
        self.emit = emit
        self.emit_sync = emit_sync
        self.log = log
        # Map of set of collect objects to list of stream names that they can be collected into
        self._declared_stream_names: dict[frozenset, list[str]] = {}

    async def open_run(self, msg: Msg):
        self.run_is_open = True
        self._run_start_uid = new_uid()
        self._interruptions_desc_uid = None  # uid for a special Event Desc.
        self._interruptions_counter = 0  # seq_num, special Event stream

        run = compose_run(uid=self._run_start_uid, event_counters=self._sequence_counters, metadata=self._md)
        doc = run.start_doc
        self._compose_descriptor = run.compose_descriptor
        self._compose_resource = run.compose_resource
        self._compose_stop = run.compose_stop
        self._compose_stream_resource = run.compose_stream_resource

        await self.emit(DocumentNames.start, doc)
        doc_logger.debug(
            "[start] document is emitted (run_uid=%r)",
            self._run_start_uid,
            extra={"doc_name": "start", "run_uid": self._run_start_uid},
        )
        await self.reset_checkpoint_state_coro()

        # Emit an Event Descriptor for recording any interruptions as Events.
        if self.record_interruptions:
            # To store the interruptions uid outside of event-model
            self._interruptions_desc_uid = new_uid()
            dk = {"dtype": "string", "shape": [], "source": "RunEngine"}
            descriptor_bundle = self._compose_descriptor(
                uid=self._interruptions_desc_uid,
                name="interruptions",
                data_keys={"interruption": dk},
            )
            self._interruptions_desc = descriptor_bundle.descriptor_doc
            self._interruptions_compose_event = descriptor_bundle.compose_event
            await self.emit(DocumentNames.descriptor, self._interruptions_desc)

        return self._run_start_uid

    async def close_run(self, msg: Msg):
        """Instruct the RunEngine to write the RunStop document

        Expected message object is::

            Msg('close_run', None, exit_status=None, reason=None)

        if *exit_stats* and *reason* are not provided, use the values
        stashed on the RE.
        """
        if not self.run_is_open:
            raise IllegalMessageSequence(
                "A 'close_run' message was received but there is no run "
                "open. If this occurred after a pause/resume, add "
                "a 'checkpoint' message after the 'close_run' message."
            )
        self.log.debug("Stopping run %r", self._run_start_uid)
        # Clear any uncleared monitoring callbacks.
        for obj, (cb, kwargs) in list(self._monitor_params.items()):  # noqa: B007
            obj.clear_sub(cb)
            del self._monitor_params[obj]
        reason = msg.kwargs.get("reason", None)
        if reason is None:
            reason = ""

        exit_status: Literal["success", "abort", "fail"] = msg.kwargs.get("exit_status", "success") or "success"
        doc = self._compose_stop(
            exit_status=exit_status,
            reason=reason,
        )
        await self.emit(DocumentNames.stop, doc)
        doc_logger.debug(
            "[stop] document is emitted (run_uid=%r)",
            self._run_start_uid,
            extra={"doc_name": "stop", "run_uid": self._run_start_uid},
        )
        await self.reset_checkpoint_state_coro()
        self.run_is_open = False
        return doc["run_start"]

    async def _prepare_stream(
        self,
        desc_key: str,
        objs_dks: dict[HasName, dict[str, DataKey]],
    ):
        # We do not have an Event Descriptor for this set
        # so one must be created.
        data_keys = {}
        config = {}
        object_keys = {}
        hints: dict[str, Any] = {}

        for obj, dks in objs_dks.items():
            maybe_update_hints(hints, obj)
            # dks is an OrderedDict. Record that order as a list.
            object_keys[obj.name] = list(dks)
            for key in dks.keys():
                dks[key]["object_name"] = obj.name
            data_keys.update(dks)
            config[obj.name] = {
                "data": self._config_values_cache[obj],
                "timestamps": self._config_ts_cache[obj],
                "data_keys": self._config_desc_cache[obj],
            }

        self._descriptors[desc_key] = self._compose_descriptor(
            desc_key,
            data_keys,
            configuration=config,
            hints=hints,
            object_keys=object_keys,
        )
        await self.emit(DocumentNames.descriptor, self._descriptors[desc_key].descriptor_doc)
        doc_logger.debug(
            "[descriptor] document emitted with name %r containing data keys %r (run_uid=%r)",
            desc_key,
            data_keys.keys(),
            self._run_start_uid,
            extra={"doc_name": "descriptor", "run_uid": self._run_start_uid, "data_keys": data_keys.keys()},
        )
        self._descriptor_objs[desc_key] = objs_dks
        if desc_key not in self._sequence_counters:
            self._sequence_counters[desc_key] = 1
            self._sequence_counters_copy[desc_key] = 1

        return (
            self._descriptors[desc_key].descriptor_doc,
            self._descriptors[desc_key].compose_event,
            list(objs_dks),
        )

    async def _ensure_cached(self, obj, collect=False):
        coros = []
        if not collect and obj not in self._describe_cache:
            coros.append(self._cache_describe(obj))
        elif collect and obj not in self._describe_collect_cache:
            coros.append(self._cache_describe_collect(obj))
        if obj not in self._config_desc_cache:
            coros.append(self._cache_describe_config(obj))
            coros.append(self._cache_read_config(obj))
        await asyncio.gather(*coros)

    async def declare_stream(self, msg):
        """Generate and emit an EventDescriptor."""
        command, no_obj, objs, kwargs, _ = msg
        stream_name = kwargs.get("name")
        assert stream_name is not None, "A stream name that is not None is required for pre-declare"

        collect = kwargs.get("collect", False)
        assert no_obj is None
        objs = frozenset(objs)
        objs_dks = {}  # {collect_object: stream_data_keys}

        await asyncio.gather(*[self._ensure_cached(obj, collect=collect) for obj in objs])
        for obj in objs:
            if collect:
                data_keys = self._describe_collect_cache[obj]
                streams_and_data_keys = self._format_datakeys_with_stream_name(
                    data_keys, message_stream_name=stream_name
                )

                # ensure that there is only one stream and it is the stream we have provided.
                assert len(streams_and_data_keys) == 1 and streams_and_data_keys[0][0] == stream_name, (
                    "`declare_stream` contained `collect=True` but  `describe_collect` did "
                    f"not return a single Dict[str, DataKey] for the passed in {stream_name}"
                )
            else:
                data_keys = self._describe_cache[obj]

            objs_dks[obj] = data_keys

        existing_stream_names = self._declared_stream_names.setdefault(objs, [])
        existing_stream_names.append(stream_name)

        return await self._prepare_stream(stream_name, objs_dks)

    async def create(self, msg):
        """
        Start bundling future obj.read() calls for an Event document.

        Expected message object is::

            Msg('create', None, name='primary')
            Msg('create', name='primary')

        Note that the `name` kwarg will be the 'name' field of the resulting
        descriptor. So descriptor['name'] = msg.kwargs['name'].

        Also note that changing the 'name' of the Event will create a new
        Descriptor document.
        """
        if self.bundling:
            raise IllegalMessageSequence(
                "A second 'create' message is not "
                "allowed until the current event "
                "bundle is closed with a 'save' or "
                "'drop' message."
            )
        self._read_cache.clear()
        self._asset_docs_cache.clear()
        self._objs_read.clear()
        self.bundling = True
        command, obj, args, kwargs, _ = msg
        try:
            self._bundle_name = kwargs["name"]
        except KeyError:
            try:
                (self._bundle_name,) = args
            except ValueError:
                raise ValueError(
                    "Msg('create') now requires a stream name, given as "
                    "Msg('create', name) or Msg('create', name=name)"
                ) from None
        if self._strict_pre_declare:
            if self._bundle_name not in self._descriptors:
                raise IllegalMessageSequence("In strict mode you must pre-declare streams.")

    async def read(self, msg, reading):
        """
        Add a reading to the open event bundle.

        Expected message object is::

            Msg('read', obj)
        """
        if self.bundling:
            obj = msg.obj
            # if the object is not in the _describe_cache, cache it
            # Note: there is a race condition between the code here
            # and in monitor() and collect(), so if you do them concurrently
            # on the same device you make obj.describe() calls multiple times.
            # As this is harmless and not an expected use case, we don't guard
            # against it. Reading multiple devices concurrently works fine.
            await self._ensure_cached(obj)

            # check that current read collides with nothing else in
            # current event
            cur_keys = set(self._describe_cache[obj].keys())
            for read_obj in self._objs_read:
                # that is, field names
                known_keys = self._describe_cache[read_obj].keys()
                if set(known_keys) & cur_keys:
                    raise ValueError(
                        f"Data keys (field names) from {obj!r} "
                        f"collide with those from {read_obj!r}. "
                        f"The colliding keys are {set(known_keys) & cur_keys}"
                    )

            # add this object to the cache of things we have read
            self._objs_read.append(obj)

            # Stash the results, which will be emitted the next time _save is
            # called --- or never emitted if _drop is called instead.
            self._read_cache.append(reading)
            # Ask the object for any resource or datum documents is has cached
            # and cache them as well. Likewise, these will be emitted if and
            # when _save is called.
            asset_docs_collected = [x async for x in maybe_collect_asset_docs(msg, obj, *msg.args, **msg.kwargs)]
            self._asset_docs_cache.extend(asset_docs_collected)

        return reading

    async def _cache_describe(self, obj):
        "Read the object's describe and cache it."
        obj = check_supports(obj, Readable)
        self._describe_cache[obj] = await maybe_await(obj.describe())

    async def _cache_describe_config(self, obj):
        "Read the object's describe_configuration and cache it."

        if isinstance(obj, Configurable):
            conf_keys = await maybe_await(obj.describe_configuration())
        else:
            conf_keys = {}

        self._config_desc_cache[obj] = conf_keys

    async def _cache_read_config(self, obj):
        "Read the object's configuration and cache it."
        if isinstance(obj, Configurable):
            conf = await maybe_await(obj.read_configuration())
        else:
            conf = {}
        config_values = {}
        config_ts = {}
        for key, val in conf.items():
            config_values[key] = val["value"]
            config_ts[key] = val["timestamp"]
        self._config_values_cache[obj] = config_values
        self._config_ts_cache[obj] = config_ts

    async def monitor(self, msg):
        """
        Monitor a signal. Emit event documents asynchronously.

        A descriptor document is emitted immediately. Then, a closure is
        defined that emits Event documents associated with that descriptor
        from a separate thread. This process is not related to the main
        bundling process (create/read/save).

        Expected message object is::

            Msg('monitor', obj, **kwargs)
            Msg('monitor', obj, name='event-stream-name', **kwargs)

        where kwargs are passed through to ``obj.subscribe()``
        """
        obj = check_supports(msg.obj, Subscribable)
        if msg.args:
            raise ValueError("The 'monitor' Msg does not accept positional arguments.")
        kwargs = dict(msg.kwargs)
        name = kwargs.pop("name", short_uid("monitor"))
        if obj in self._monitor_params:
            raise IllegalMessageSequence(f"A 'monitor' message was sent for {obj} which is already monitored")

        await self._ensure_cached(obj)

        stream_bundle = await self._prepare_stream(name, {obj: self._describe_cache[obj]})
        compose_event = stream_bundle[1]

        def emit_event(readings: Optional[dict[str, Reading]] = None, *args, **kwargs):
            if readings is not None:
                # We were passed something we can use, but check no args or kwargs
                assert not args and not kwargs, (
                    "If subscribe callback called with readings, args and kwargs are not supported."
                )
            else:
                # Ignore the inputs. Use this call as a signal to call read on the
                # object, a crude way to be sure we get all the info we need.
                readable_obj = check_supports(obj, Readable)  # type: ignore
                readings = readable_obj.read()  # type: ignore
                assert not inspect.isawaitable(readings), (
                    f"{readable_obj} has async read() method and the callback "
                    "passed to subscribe() was not called with Dict[str, Reading]"
                )
            data, timestamps = _rearrange_into_parallel_dicts(readings)
            doc = compose_event(
                data=data,
                timestamps=timestamps,
            )
            self.emit_sync(DocumentNames.event, doc)

        self._monitor_params[obj] = emit_event, kwargs
        # TODO: deprecate **kwargs when Ophyd.v2 is available
        obj.subscribe(emit_event, **kwargs)

    def record_interruption(self, content):
        """
        Emit an event in the 'interruptions' event stream.

        If we are not inside a run or if self.record_interruptions is False,
        nothing is done.
        """
        if self._interruptions_desc_uid is not None:
            # We are inside a run and self.record_interruptions is True.
            doc = self._interruptions_compose_event(
                data={"interruption": content},
                timestamps={"interruption": ttime.time()},
            )
            self._interruptions_counter += 1
            self.emit_sync(DocumentNames.event, doc)

    def rewind(self):
        self._sequence_counters.clear()
        self._sequence_counters.update(self._sequence_counters_copy)
        # make sure we do not forget about streams we roll back to the
        # very beginning of
        for desc_key in self._descriptor_objs:
            if desc_key not in self._sequence_counters:
                self._sequence_counters[desc_key] = 1
                self._sequence_counters_copy[desc_key] = 1

        # This is needed to 'cancel' an open bundling (e.g. create) if
        # the pause happens after a 'checkpoint', after a 'create', but
        # before the paired 'save'.
        self.bundling = False

    async def unmonitor(self, msg):
        """
        Stop monitoring; i.e., remove the callback emitting event documents.

        Expected message object is::

            Msg('unmonitor', obj)
        """
        obj = check_supports(msg.obj, Subscribable)
        if obj not in self._monitor_params:
            raise IllegalMessageSequence(f"Cannot 'unmonitor' {obj}; it is not being monitored.")
        cb, kwargs = self._monitor_params[obj]
        obj.clear_sub(cb)
        del self._monitor_params[obj]
        await self.reset_checkpoint_state_coro()

    async def save(self, msg):
        """Save the event that is currently being bundled

        Create and emit an Event document containing the data read from devices
        in self._objs_read. Emit any Resource and Datum documents cached by
        those devices before emitting the Event document. If this is the first
        Event of its stream then create and emit the Event Descriptor document
        before emitting Resource, Datum, and Event documents.

        Expected message object is::

            Msg('save')
        """
        if not self.bundling:
            raise IllegalMessageSequence(
                "A 'create' message must be sent, to "
                "open an event bundle, before that "
                "bundle can be saved with 'save'."
            )

        # Short-circuit if nothing has been read. (Do not create empty Events.)
        if not self._objs_read:
            self.bundling = False
            self._bundle_name = None
            return
        # The Event Descriptor is uniquely defined by the set of objects
        # read in this Event grouping.
        objs_read = frozenset(self._objs_read)

        # Event Descriptor key
        desc_key = self._bundle_name

        # This is a separate check because it can be reset on resume.
        self.bundling = False
        self._bundle_name = None

        (
            descriptor_doc,
            compose_event,
            _,
        ) = self._descriptors.get(desc_key, (None, None, None))
        d_objs = self._descriptor_objs.get(desc_key, None)

        objs_dks = {}

        # we do not have the descriptor cached, make it
        if descriptor_doc is None or d_objs is None:
            # use the dequeue not the set to preserve order
            for obj in self._objs_read:
                await self._ensure_cached(obj, collect=isinstance(obj, Collectable))
                objs_dks[obj] = self._describe_cache[obj]

            descriptor_doc, compose_event, d_objs = await self._prepare_stream(desc_key, objs_dks)

        # do have the descriptor cached
        elif frozenset(d_objs) != objs_read:
            raise RuntimeError(f"Mismatched objects read, expected {frozenset(d_objs)!s}, got {objs_read!s}")

        # Resource and Datum documents
        indices_generated = await self._pack_external_assets(self._asset_docs_cache, message_stream_name=desc_key)
        if indices_generated > 1:
            raise RuntimeError(
                "Received multiple indices in a `stream_datum` document for one event, "
                ' during a `read()` `save()`. `stream_datum` should have indices {"start": n, "stop": n+1} '
                "in a `read()` `save()`."
            )

        # Merge list of readings into single dict.
        readings = {k: v for d in self._read_cache for k, v in d.items()}
        data, timestamps = _rearrange_into_parallel_dicts(readings)
        # Mark all externally-stored data as not filled so that consumers
        # know that the corresponding data are identifiers, not dereferenced
        # data.
        filled = {
            k: False
            for k, v in self._descriptors[desc_key].descriptor_doc["data_keys"].items()
            if "external" in v and v["external"] != "STREAM:"
        }
        event_doc = compose_event(
            data=data,
            timestamps=timestamps,
            filled=filled,
        )
        await self.emit(DocumentNames.event, event_doc)
        doc_logger.debug(
            "[event] document emitted with data keys %r (run_uid=%r)",
            data.keys(),
            self._run_start_uid,
            extra={"doc_name": "event", "run_uid": self._run_start_uid, "data_keys": data.keys()},
        )

    def clear_monitors(self):
        for obj, (cb, kwargs) in list(self._monitor_params.items()):  # noqa: B007
            try:
                obj.clear_sub(cb)
            except Exception:
                self.log.exception("Failed to stop monitoring %r.", obj)
            else:
                del self._monitor_params[obj]

    def reset_checkpoint_state(self):
        # Keep a safe separate copy of the sequence counters to use if we
        # rewind and retake some data points.

        for key, counter in list(self._sequence_counters.items()):
            self._sequence_counters_copy[key] = counter

    async def reset_checkpoint_state_coro(self):
        self.reset_checkpoint_state()

    async def suspend_monitors(self):
        for obj, (cb, kwargs) in self._monitor_params.items():  # noqa: B007
            obj.clear_sub(cb)

    async def restore_monitors(self):
        for obj, (cb, kwargs) in self._monitor_params.items():
            obj.subscribe(cb, **kwargs)

    async def clear_checkpoint(self, msg):
        self._sequence_counters_copy.clear()

    async def drop(self, msg):
        """Drop the event that is currently being bundled

        Expected message object is::

            Msg('drop')
        """
        if not self.bundling:
            raise IllegalMessageSequence(
                "A 'create' message must be sent, to "
                "open an event bundle, before that "
                "bundle can be dropped with 'drop'."
            )

        self.bundling = False
        self._bundle_name = None
        self.log.debug("Dropped open event bundle")

    async def kickoff(self, msg):
        """Start a flyscan object.

        Expected message object is:

        If `flyer_object` has a `kickoff` function that takes no arguments::

            Msg('kickoff', flyer_object)
            Msg('kickoff', flyer_object, group=<name>)

        If *flyer_object* has a ``kickoff`` function that takes
        ``(start, stop, steps)`` as its function arguments::

            Msg('kickoff', flyer_object, start, stop, step)
            Msg('kickoff', flyer_object, start, stop, step, group=<name>)

        """
        self._uncollected.add(msg.obj)

    def _format_datakeys_with_stream_name(
        self,
        describe_collect_dict: Union[dict[str, DataKey], dict[str, dict[str, DataKey]]],
        message_stream_name: Optional[str] = None,
    ) -> list[tuple[str, dict[str, DataKey]]]:
        """
        Check if the dictionary returned by describe collect is a dict
            `{str: DataKey}` or a `{str: {str: DataKey}}`.
        If a `message_stream_name` is passed then return a singleton list of the form of
            `{message_stream_name: describe_collect_dict}.items()`.
        If the `message_stream_name` is None then return the `describe_collect_dict.items()`.
        """

        def _contains_message_stream_name(
            describe_collect_dict: Union[Any, dict[str, Any]],
        ) -> bool:  # TODO: change to TypeGuard[dict[str, dict[str, DataKey]]] after python 3.9
            return isinstance(describe_collect_dict, dict) and all(
                _describe_collect_dict_is_valid(v) for v in describe_collect_dict.values()
            )

        if describe_collect_dict:
            if _describe_collect_dict_is_valid(describe_collect_dict):
                # TODO: remove cast after python 3.9 is no longer supported
                flat_describe_collect_dict = cast(dict[str, DataKey], describe_collect_dict)
                return [(message_stream_name or "primary", flat_describe_collect_dict)]
            # Validate that all of the values nested values are DataKeys
            elif _contains_message_stream_name(describe_collect_dict):
                # We have Dict[str, Dict[str, DataKey]] so return its items
                # TODO: remove cast after python 3.9 is no longer supported
                nested_describe_collect_dict = cast(dict[str, dict[str, DataKey]], describe_collect_dict)
                if message_stream_name and list(nested_describe_collect_dict) != [message_stream_name]:
                    # The collect contained a name and describe_collect returned a Dict[str, Dict[str, DataKey]],
                    # this is only acceptable if the only key in the parent dict is message_stream_name

                    raise RuntimeError(
                        f"Expected a single stream {message_stream_name!r}, got {describe_collect_dict}"
                    )
                return list(nested_describe_collect_dict.items())
            else:
                raise RuntimeError(
                    f"Invalid describe_collect return: {describe_collect_dict} when collect "
                    f"was called on {message_stream_name}"
                )
        else:
            # Empty dict, could be either but we don't care
            return []

    async def _cache_describe_collect(self, obj: Collectable):
        "Read the object's describe and cache it."
        obj = check_supports(obj, Collectable)
        c: Union[dict[str, DataKey], dict[str, dict[str, DataKey]]] = await maybe_await(obj.describe_collect())
        self._describe_collect_cache[obj] = c

    async def _describe_collect(self, collect_object: Flyable):
        """Read an object's describe_collect and cache it.

        Read describe collect for a collect_object and ensure it is cached in the
        _describe_collect_cache. This is required for scans of a single collect object
        where the data structure is doubly nested. In this case calling
        describe_collect on the object returns a data structure like so:

            {
                "stream1" : {"stream1-pv1":{data_keys}, "stream1-pv2" :{data_keys}},
                "stream2" : {"stream2-pv1":{data_keys}, "stream2-pv2" :{data_keys}}
            }

        Single nested data keys should be rejected since they are new style, and are
        collected under one stream. They should be pre-declared with declare-stream,
        prior to collecting. The describe_collect on this object returns a data
        structure like so:
            {
                "stream1-pv1": {data_keys},
                "stream1-pv1": {data_keys}
            }

        """
        await self._ensure_cached(collect_object, collect=True)

        describe_collect = self._describe_collect_cache[collect_object]
        describe_collect_items = self._format_datakeys_with_stream_name(describe_collect)

        local_descriptors: dict[frozenset[str], ComposeDescriptorBundle] = {}

        # Check that singly nested stuff should have been pre-declared
        def is_data_key(obj: Any) -> bool:
            return isinstance(obj, dict) and {"dtype", "shape", "source"}.issubset(frozenset(obj.keys()))

        assert all(not is_data_key(value) for value in describe_collect.values()), (
            "Single nested data keys should be pre-declared"
        )

        # Make sure you can't use identical data keys in multiple streams
        # Data structure is assumed to be dict[stream_name, dictionary of key -> data_key]
        duplicates: dict[str, dict[str, DataKey]] = defaultdict(dict)
        for (
            (stream1, data_keys1),
            (stream2, data_keys2),
        ) in combinations(describe_collect_items, 2):
            for common_key in data_keys1.keys() & data_keys2.keys():
                duplicates[stream1][common_key] = data_keys1[common_key]
                duplicates[stream2][common_key] = data_keys2[common_key]
        if len(duplicates) > 0:
            raise RuntimeError(
                "Can't use identical data keys in multiple streams:",
                f"Data keys: {list(duplicates.values())}",
                f"streams: {duplicates.keys()}",
            )

        for stream_name, stream_data_keys in describe_collect_items:
            if stream_name not in self._descriptor_objs or (
                collect_object not in self._descriptor_objs[stream_name]
            ):
                await self._prepare_stream(stream_name, {collect_object: stream_data_keys})
            else:
                objs_read = self._descriptor_objs[stream_name]
                if stream_data_keys != objs_read[collect_object]:
                    raise RuntimeError(
                        f"Mismatched objects read, expected {stream_data_keys!s}, got {objs_read!s}"
                    )
            local_descriptors[frozenset(stream_data_keys)] = self._descriptors[stream_name]

        self._local_descriptors[collect_object] = local_descriptors

    async def _pack_seq_nums_into_stream_datum(
        self, doc: StreamDatum, message_stream_name: str, stream_datum_previous_indices_difference: int
    ) -> int:
        if doc["seq_nums"] != StreamRange(start=0, stop=0):
            raise EventModelValueError(
                f"Received `seq_nums` {doc['seq_nums']} in stream {doc['stream_resource']} "
                "during `collect()` or `describe_collect()`. `seq_nums` should be None or "
                "`StreamRange(start=0, stop=0)` on `ComposeStreamDatum` when used with the "
                "run engine."
            )

        indices_difference = doc["indices"]["stop"] - doc["indices"]["start"]

        if (
            stream_datum_previous_indices_difference
            and stream_datum_previous_indices_difference != indices_difference
        ):
            raise EventModelValueError(
                f"Received `indices` {doc['indices']} during `collect()` these are of a different "
                f"width `{indices_difference}` than other detectors in the same collect() or save()."
            )

        current_seq_counter = self._sequence_counters[message_stream_name]
        doc["seq_nums"] = StreamRange(start=current_seq_counter, stop=current_seq_counter + indices_difference)

        if doc["stream_resource"] not in self._stream_resource_data_keys:
            raise RuntimeError(
                f"Receieved a `steam_datum` referring to an unknown stream resource {doc['stream_resource']}"
            )

        return indices_difference

    # message strem name here?
    async def _pack_external_assets(
        self, asset_docs: Iterable[Union[Asset, StreamAsset]], message_stream_name: Optional[str]
    ):
        """Packs some external asset documents with relevant information from the run."""

        stream_datum_previous_indices_difference = 0
        data_keys_received = set()
        descriptor_doc = None
        external_data_keys = None

        if message_stream_name:
            descriptor_doc = self._descriptors[message_stream_name].descriptor_doc
            external_data_keys = self.get_external_data_keys(descriptor_doc["data_keys"])

        for name, doc in asset_docs:
            if name == DocumentNames.resource.value:
                resource_doc = cast(Resource, doc)
                if self._run_start_uid is None:
                    raise RuntimeError("Run start uid was never set before Resource document was emitted.")
                resource_doc["run_start"] = self._run_start_uid
            elif name == DocumentNames.stream_resource.value:
                stream_resource_doc = cast(StreamResource, doc)
                if self._run_start_uid is None:
                    raise RuntimeError("Run start uid was never set before StreamResource document was emitted.")
                stream_resource_doc["run_start"] = self._run_start_uid

                if stream_resource_doc["uid"] in self._stream_resource_data_keys:
                    raise RuntimeError(f"Received `stream_resource` with uid {stream_resource_doc['uid']} twice.")

                self._stream_resource_data_keys[stream_resource_doc["uid"]] = stream_resource_doc["data_key"]

                if not external_data_keys or stream_resource_doc["data_key"] not in external_data_keys:
                    raise RuntimeError(
                        f"Receieved a `stream_resource` with data_key {stream_resource_doc['data_key']} "
                        f"that is not in the descriptor 'STREAM:' data_keys {external_data_keys}"
                    )

            elif name == DocumentNames.stream_datum.value:
                stream_datum_doc = cast(StreamDatum, doc)
                if stream_datum_doc["descriptor"]:
                    raise RuntimeError(
                        f"Received a `stream_datum` {stream_datum_doc['uid']} with a `descriptor` uid already "
                        f"filled in, with the value {stream_datum_doc['descriptor']} this should be an empty "
                        f"string."
                    )
                if not descriptor_doc:
                    raise RuntimeError(f"`descriptor` not made for stream {message_stream_name}.")
                data_keys_received.add(self._stream_resource_data_keys[stream_datum_doc["stream_resource"]])

                stream_datum_doc["descriptor"] = descriptor_doc["uid"]
                if message_stream_name is None:
                    raise RuntimeError(
                        "`message_stream_name` was never set before StreamDatum document was emitted."
                    )
                stream_datum_previous_indices_difference = await self._pack_seq_nums_into_stream_datum(
                    stream_datum_doc,
                    message_stream_name,
                    stream_datum_previous_indices_difference,
                )
            elif name == DocumentNames.datum.value:
                ...
            else:
                raise RuntimeError(
                    f"Tried to emit an external asset {name}, acceptable external assets are "
                    "`resource`, `stream_resource`, `datum`, or `stream_datum`"
                )

            await self.emit(DocumentNames(name), doc)

            doc_logger.debug(
                "[%s] document emitted %r",
                name,
                doc,
                extra={"doc_name": name, "run_uid": self._run_start_uid, "doc": doc},
            )

        # Check we have a stream_datum for each external data_key in the descriptor
        if descriptor_doc and data_keys_received and set(external_data_keys) != data_keys_received:  # type: ignore
            raise RuntimeError(
                f"Received `stream_datum` for each of the data_keys {data_keys_received}, "  # type: ignore
                f"expected `stream_datum` for each of the data_keys {set(external_data_keys)}."  # type: ignore
            )

        return stream_datum_previous_indices_difference

    def get_external_data_keys(self, data_keys: dict[str, DataKey]) -> list[str]:
        """Get the external data keys from the descriptor data_keys dictionary"""
        return [x for x in data_keys if ("external" in data_keys[x] and data_keys[x]["external"] == "STREAM:")]

    async def _collect_events(
        self,
        collect_obj: EventCollectable,
        local_descriptors,
        return_payload: bool,
        message_stream_name: Optional[str],
    ):
        payload = []
        pages: dict[frozenset[str], list[Event]] = defaultdict(list)

        if message_stream_name:
            compose_event = self._descriptors[message_stream_name].compose_event
            data_keys = self._descriptors[message_stream_name].descriptor_doc["data_keys"]
            objs_read = frozenset(data_keys.keys())

        async for partial_event in iterate_maybe_async(collect_obj.collect()):
            if return_payload:
                payload.append(partial_event)

            if not message_stream_name:
                objs_read = frozenset(partial_event["data"])
                compose_event = local_descriptors[objs_read].compose_event
                data_keys = local_descriptors[objs_read].descriptor_doc["data_keys"]
                assert frozenset(data_keys.keys()) == objs_read

            if [x for x in self.get_external_data_keys(data_keys) if x in partial_event["data"]]:
                raise RuntimeError("Received an event containing data for external data keys.")

            # is there a way to generalise the keys?
            if "filled" in partial_event.keys():
                event = compose_event(
                    data=partial_event["data"],
                    timestamps=partial_event["timestamps"],
                    filled=partial_event["filled"],
                )
            else:
                event = compose_event(data=partial_event["data"], timestamps=partial_event["timestamps"])

            pages[objs_read].append(event)

        for event_list in pages.values():
            await self.emit(DocumentNames.event_page, pack_event_page(*event_list))
            doc_logger.debug(
                "[event_page] document is emitted for descriptors (run_uid=%r)",
                self._run_start_uid,
                extra={"doc_name": "event_page", "run_uid": self._run_start_uid},
            )
        return payload

    async def _collect_event_pages(
        self,
        collect_obj: EventPageCollectable,
        local_descriptors,
        return_payload: bool,
        message_stream_name: Optional[str],
    ):
        payload = []

        if message_stream_name:
            compose_event_page = self._descriptors[message_stream_name].compose_event_page
            data_keys = self._descriptors[message_stream_name].descriptor_doc["data_keys"]

        async for ev_page in iterate_maybe_async(collect_obj.collect_pages()):
            if return_payload:
                payload.append(ev_page)

            if not message_stream_name:
                objs_read = frozenset(ev_page["data"])
                compose_event_page = local_descriptors[objs_read].compose_event_page
                data_keys = local_descriptors[objs_read].descriptor_doc["data_keys"]

            if [x for x in self.get_external_data_keys(data_keys) if x in ev_page["data"]]:
                raise RuntimeError("Received an event_page containing data for external data keys.")

            ev_page = compose_event_page(data=ev_page["data"], timestamps=ev_page["timestamps"])
            doc_logger.debug(
                "[event_page] document is emitted with data keys %r (run_uid=%r)",
                ev_page["data"].keys(),
                ev_page["uid"],
                extra={
                    "doc_name": "event_page",
                    "run_uid": self._run_start_uid,
                    "data_keys": ev_page["data"].keys(),
                },
            )

            await self.emit(DocumentNames.event_page, ev_page)
        return payload

    async def collect(self, msg: Msg):
        """
        Collect data cached by a flyer and emit documents.

        Expect message object is
            Msg('collect',  collect_obj,  collect_obj_2, ..., stream=True,
                return_payload=True, name='stream_name')

        Where there must be at least one collect object. If multiple are used
        they must obey the WritesStreamAssets protocol.
        """
        stream_name: Optional[str] = None

        if not self.run_is_open:
            # sanity check -- 'kickoff' should catch this and make this
            # code path impossible
            raise IllegalMessageSequence("A 'collect' message was sent but no run is open.")

        # If stream is True, run 'event' subscription per document.
        # If stream is False, run 'event_page' subscription once.
        # Stream is True is no longer supported
        stream = msg.kwargs.get("stream", False)
        if stream is True:
            raise RuntimeError(
                "Collect now emits EventPages (stream=False), "
                "so emitting Events (stream=True) is no longer supported"
            )

        # If True, accumulate all the Events in memory and return them at the
        # end, providing the plan access to the Events. If False, do not
        # accumulate, and return None.
        return_payload = msg.kwargs.get("return_payload", True)

        # Get a list of the collectable objects from the message obj and args
        collect_objects = [check_supports(obj, Collectable) for obj in (msg.obj,) + msg.args]

        # Get references to get_index methods if we have more than one collect object
        # raise error if collect_objects don't obey WritesStreamAssests protocol
        indices: list[Callable[[], SyncOrAsync[int]]] = []
        if len(collect_objects) > 1:
            indices = [check_supports(obj, WritesStreamAssets).get_index for obj in collect_objects]

        # Warn for page collectable support
        for obj in collect_objects:
            if isinstance(obj, EventCollectable) and isinstance(obj, EventPageCollectable):
                doc_logger.warn(
                    "collect() was called for a device %r which is both EventCollectable "
                    "and EventPageCollectable. Using device.collect_pages().",
                    obj.name,
                )
            self._uncollected.discard(obj)

        # Get the provided message stream name for singly nested scans
        message_stream_name: Optional[str] = msg.kwargs.get("name", None)

        # Retrive the stream names from pre-declared streams
        declared_stream_names = self._declared_stream_names.get(frozenset(collect_objects), [])

        # If a stream name was provided in the message, check the stream has been declared
        # If one was not provided, but a single stream has been declared, then use that stream.
        if message_stream_name:
            assert message_stream_name in declared_stream_names, (
                "If a message stream name is provided declare stream needs to be called first."
            )
            stream_name = message_stream_name
        elif declared_stream_names:
            assert len(frozenset(declared_stream_names)) == 1  # Allow duplicate declarations
            stream_name = declared_stream_names[0]

        # If there is not a stream then we should be using an old-style doubly nested
        # and we need to describe_collect and prepare the nested streams.
        if not stream_name:
            if frozenset(collect_objects) not in self._local_descriptors or (
                collect_objects[0] not in self._local_descriptors
            ):
                if len(collect_objects) > 1:
                    raise IllegalMessageSequence(
                        "If collecting multiple objects you must predeclare a stream for all "
                        "the objects first and provide the stream name"
                    )
                else:
                    await self._describe_collect(collect_objects[0])

        # Get the indicies from the collect objects
        coros = [maybe_await(get_index()) for get_index in indices]
        if coros:
            # There is more than one collect object, so collect up to a minimum index
            min_index = min(await asyncio.gather(*coros))
        else:
            # There is only one collect object, so don't pass an index down
            min_index = None

        collected_asset_docs = [
            x
            for obj in collect_objects
            async for x in maybe_collect_asset_docs(
                msg,
                obj,
                index=min_index,
            )
        ]

        indices_difference = await self._pack_external_assets(
            collected_asset_docs, message_stream_name=stream_name
        )

        # Make event pages for an object which is EventCollectable or EventPageCollectable
        # objects that are EventCollectable will now group the Events and Emit an Event Page
        if len(collect_objects) == 1 and not isinstance(collect_objects[0], WritesStreamAssets):
            local_descriptors: dict[frozenset[str], ComposeDescriptorBundle] = {}
            collect_obj = collect_objects[0]

            # If the single collect object is singly nested, gather descriptors
            if collect_obj not in self._local_descriptors:
                if stream_name is None:
                    raise RuntimeError(
                        "Expected stream_name to be set before calling collect() on a singly nested device."
                    )
                objs = self._descriptor_objs[stream_name]
                data_keys = objs[collect_obj]
                local_descriptors[frozenset(data_keys)] = self._descriptors[stream_name]
                self._local_descriptors[collect_obj] = local_descriptors

            local_descriptors = self._local_descriptors[collect_obj]

            if isinstance(collect_obj, EventPageCollectable):
                payload = await self._collect_event_pages(
                    collect_obj, local_descriptors, return_payload, stream_name
                )
                # TODO: check that event pages have same length as indices_difference
            elif isinstance(collect_obj, EventCollectable):
                payload = await self._collect_events(collect_obj, local_descriptors, return_payload, stream_name)
                # TODO: check that events have same length as indices_difference
            else:
                return_payload = False
                if not stream_name:
                    raise RuntimeError(
                        "A `collect` message on a device that isn't EventCollectable or EventPageCollectable "
                        "requires a `name=stream_name` argument"
                    )
                # Since there are no events or event_pages incrementing the sequence counter, we do it ourselves.
                self._sequence_counters[stream_name] += indices_difference

            if return_payload:
                return payload

        else:
            # Since there are no events or event_pages incrementing the sequence counter, we do it ourselves.
            self._sequence_counters[stream_name] += indices_difference

    async def backstop_collect(self):
        for obj in list(self._uncollected):
            try:
                await self.collect(Msg("collect", obj))
            except Exception:
                self.log.exception("Failed to collect %r.", obj)

    async def configure(self, msg):
        """Configure an object

        Expected message object is ::

            Msg('configure', object, *args, **kwargs)

        which results in this call ::

            object.configure(*args, **kwargs)
        """
        obj = msg.obj
        await self._cache_read_config(obj)
        # Invalidate any event descriptors that include this object.
        # New event descriptors, with this new configuration, will
        # be created for any future event documents.
        for name in list(self._descriptors):
            obj_set = self._descriptor_objs[name]
            if obj in obj_set:
                del self._descriptors[name]
                await self._prepare_stream(name, obj_set)
                continue
